"""Manage VLANS on HPCOM7 devices.
"""
from pyhpecw7.features.errors import LengthOfStringError, VlanIDError
from pyhpecw7.utils.xml.lib import *


class Vlan(object):
    """This class is used to get data and configure a specific VLAN.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        vlanid (str): VLAN ID

    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        vlanid (str): VLAN ID

    """
    def __init__(self, device, vlanid=None):
        self.device = device
        if vlanid:
            self.vlanid = vlanid

        # dictionary to XML tag mappings
        self.vlan_key_map = {
            'vlanid': 'ID',
            'name': 'Name',
            'descr': 'Description'
        }

    def gen_top(self):
        E = data_element_maker()
        top = E.top(
            E.VLAN(
                E.VLANs(
                    E.VLANID()
                )
            )
        )

        return top

    def get_vlan_list(self):
        """Get a list of VLAN IDs that exist on the switch.

        Returns:
            It returns a list of VLAN IDs as strings.
        """
        top = self.gen_top()
        nc_get_reply = self.device.get(('subtree', top))
        vlans_xml = findall_in_data('ID', nc_get_reply.data_ele)
        vlans = [vlan.text for vlan in vlans_xml]

        return vlans

    def get_config(self):
        """Gets current configuration for a given VLAN ID

        Args:
            vlanid (str): REQUIRED - VLAN ID

        Returns:
            This returns a dictionary with the following
                k/v pairs:

                :vlanid (str): VLAN ID of the vlan requested
                :name (str): configured name of the vlan
                :descr (str): configured descr of the vlan

            It returns an empty dictionary if the vlan does not exist
        """
        top = self.gen_top()
        vlan_id_ele = find_in_data('VLANID', top)
        vlan_id_ele.append(data_element_maker().ID(self.vlanid))

        nc_get_reply = self.device.get(('subtree', top))
        vlan_config = data_elem_to_dict(nc_get_reply.data_ele, self.vlan_key_map)

        return vlan_config

    def remove(self, stage=False):
        """Stage or execute XML object for VLAN removal and send to staging

        Args:
            stage (bool): whether to stage the command or execute immediately

        Returns:
            True if stage=True and successfully staged
            etree.Element XML response if immediate execution
        """
        config = self._build_config(state='absent')
        if stage:
            return self.device.stage_config(config, 'edit_config')
        else:
            return self.device.edit_config(config)

    def build(self, stage=False, **vlan):
        """Stage or execute XML object for VLAN configuration and send to staging

        Args:
            stage (bool): whether to stage the command or execute immediately
            vlan: see Keyword Args

        Keyword Args:
            name (str): OPTIONAL - VLAN name
            descr (str): OPTOINAL - VLAN description

        Returns:
            True if stage=True and successfully staged
            etree.Element XML response if immediate execution
        """
        config = self._build_config(state='present', **vlan)
        if stage:
            return self.device.stage_config(config, 'edit_config')
        else:
            return self.device.edit_config(config)

    def _build_config(self, state, **vlan):
        """Build XML object for VLAN configuration

        Args:
            state (str): OPTIONAL must be "present" or "absent"
                DEFAULT is "present"
            vlan: see Keyword Args

            Keyword Args:
                name (str): OPTIONAL - VLAN name
                descr (str): OPTOINAL - VLAN description

        Returns:
            XML object for VLAN configuration
        """
        if state == 'present':
            operation = 'merge'
        elif state == 'absent':
            operation = 'delete'

        EC = nc_element_maker()
        E = config_element_maker()

        vlan['vlanid'] = self.vlanid
        config = EC.config(
            E.top(
                E.VLAN(
                    E.VLANs(
                        E.VLANID(
                            *config_params(vlan, self.vlan_key_map)
                        )
                    ),
                    **operation_kwarg(operation)
                )
            )
        )

        return config

    def param_check(self, **vlan):
        """Basic param validation for vlan

        Args:
            state (str): REQUIRED must be "present" or "absent"
            vlan: see Keyword Args

            Keyword Args:
                vlanid (str): OPTIONAL - VLAN ID
                name (str): OPTIONAL - VLAN name
                descr (str): OPTIONAL - VLAN description

        """
        try:
            vlanid = int(self.vlanid)
        except ValueError:
            raise VlanIDError
        if vlanid < -1 or vlanid > 4094:
            raise VlanIDError

        descr = vlan.get('descr')
        if descr and len(descr) > 254:
            raise LengthOfStringError("'descr'")

        name = vlan.get('name')
        if name and len(name) > 32:
            raise LengthOfStringError("'name'")



import unittest
import mock

from pyhpecw7.features.install_os import InstallOs
from .base_feature_test import BaseFeatureCase


class InstallOsTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self, mock_device):
        self.device = mock_device
        self.ios = InstallOs(self.device)

    def test_get_config(self):
        expected_get, get_reply = self.xml_get_and_reply('install_os')
        self.device.get.return_value = get_reply

        expected = {'current': {'boot': '5930-cmw710-boot-e2415.bin', 'system': '5930-cmw710-system-e2415.bin'}, 'startup-primary': {'boot': '5930-cmw710-boot-e2415.bin', 'system': '5930-cmw710-system-e2415.bin'}}

        result = self.ios.get_config()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    def test_build_ipe(self):
        expected_call = self.read_action_xml('install_os_ipe')
        result = self.ios.build('ipe', ipe='ipe_name', delete_ipe=True)
        self.assert_action_request(expected_call)

        result = self.ios.build('ipe', ipe='ipe_name', delete_ipe=True, stage=True)
        self.assert_stage_request(expected_call, 'action')

    def test_build_boot_sys(self):
        expected_call = self.read_action_xml('install_os_bootsys')
        result = self.ios.build('bootsys', boot='boot_image', system='sys_image')
        self.assert_action_request(expected_call)

        result = self.ios.build('bootsys', boot='boot_image', system='sys_image', stage=True)
        self.assert_stage_request(expected_call, 'action')


if __name__ == "__main__":
    unittest.main()


import unittest
import mock

from pyhpecw7.features.vrrp import VRRP
from .base_feature_test import BaseFeatureCase

IFACE_NAME = 'vlan100'
VRID = '100'

class VRRPTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self, mock_device):
        self.device = mock_device
        self.vrrp = VRRP(self.device, IFACE_NAME, VRID)

    def test_get_groups(self):
        expected_cmd = 'display vrrp interface vlan100 verbose'
        RAW_OUT = """IPv4 Virtual Router Information:
 Running mode      : Standard
 Total number of virtual routers on interface Vlan-interface100 : 1
   Interface Vlan-interface100
     VRID           : 100                 Adver Timer  : 100
     Admin Status   : Up                  State        : Initialize
     Config Pri     : 100                 Running Pri  : 100
     Preempt Mode   : Yes                 Delay Time   : 0
     Auth Type      : None
     Virtual IP     : 100.100.100.1
     Master IP      : 0.0.0.0
"""
        self.device.cli_display.return_value = RAW_OUT

        result = self.vrrp.get_vrrp_groups()
        expected = ['100']

        self.assertEqual(result, expected)
        self.device.cli_display.assert_called_with(expected_cmd)

    def test_get_auth(self):
        expected_cmd = 'display current-configuration interface vlan100 | inc "vrid 100 auth"'
        RAW_OUT = """<HP1>display current-configuration interface vlan100 | inc "vrid 100 auth"
 vrrp vrid 100 authentication-mode simple cipher $c$3$ePQsjr4juyIoqhrNpWXUhRx0JQr220UaQj8=

"""
        self.device.cli_display.return_value = RAW_OUT

        result = self.vrrp.get_auth_type()
        expected = {'key_type': 'cipher', 'auth_mode': 'simple', 'key': '$c$3$ePQsjr4juyIoqhrNpWXUhRx0JQr220UaQj8='}

        self.assertEqual(result, expected)
        self.device.cli_display.assert_called_with(expected_cmd)

    def test_remove(self):
        expected_cmds = ['interface vlan100', 'undo vrrp vrid 100', '\n']

        result = self.vrrp.remove()
        self.device.cli_config.assert_called_with(expected_cmds)
        self.assertEqual(result, self.device.cli_config.return_value)

        result = self.vrrp.remove(stage=True)
        self.device.stage_config.assert_called_with(expected_cmds, 'cli_config')
        self.assertEqual(result, self.device.stage_config.return_value)

    def test_shutdown(self):
        expected_cmds = ['interface vlan100', 'vrrp vrid 100 shutdown', '\n']

        result = self.vrrp.shutdown()
        self.device.cli_config.assert_called_with(expected_cmds)
        self.assertEqual(result, self.device.cli_config.return_value)

        result = self.vrrp.shutdown(stage=True)
        self.device.stage_config.assert_called_with(expected_cmds, 'cli_config')
        self.assertEqual(result, self.device.stage_config.return_value)

    def test_undoshutdown(self):
        expected_cmds = ['interface vlan100', 'undo vrrp vrid 100 shutdown', '\n']

        result = self.vrrp.undoshutdown()
        self.device.cli_config.assert_called_with(expected_cmds)
        self.assertEqual(result, self.device.cli_config.return_value)

        result = self.vrrp.undoshutdown(stage=True)
        self.device.stage_config.assert_called_with(expected_cmds, 'cli_config')
        self.assertEqual(result, self.device.stage_config.return_value)

    @mock.patch.object(VRRP, 'get_auth_type')
    @mock.patch.object(VRRP, 'get_vrrp_groups')
    def test_get_config(self, mock_groups, mock_auth):
        mock_auth.return_value = {'key_type': 'cipher', 'auth_mode': 'simple', 'key': '$c$3$TMGInQzbelfMG2TrL1ESE+mkxEWfT+UbkOw='}
        mock_groups.return_value = ['100']

        self.vrrp._existing_all = [{'key_type': 'cipher', 'admin': 'Up', 'vrid': '100', 'vip': '100.100.100.1', 'auth_mode': 'simple', 'priority': '100', 'preempt': 'no', 'key': '$c$3$TMGInQzbelfMG2TrL1ESE+mkxEWfT+UbkOw='}]

        expected = {'key_type': 'cipher', 'admin': 'Up', 'vrid': '100', 'vip': '100.100.100.1', 'auth_mode': 'simple', 'priority': '100', 'preempt': 'no', 'key': '$c$3$TMGInQzbelfMG2TrL1ESE+mkxEWfT+UbkOw='}

        result = self.vrrp.get_config()
        self.assertEqual(result, expected)

    def test_build(self):
        kwargs = dict(vip='100', auth_mode='md5', key_type='plain', key='secret', priority='10', preempt='yes')
        expected_cmds = ['interface vlan100', 'vrrp vrid 100 virtual-ip 100', 'vrrp vrid 100 priority 10', 'vrrp vrid 100 preempt-mode', 'vrrp vrid 100 authentication-mode md5 plain secret', '\n']

        result = self.vrrp.build(**kwargs)
        self.device.cli_config.assert_called_with(expected_cmds)
        self.assertEqual(result, self.device.cli_config.return_value)

        result = self.vrrp.build(stage=True, **kwargs)
        self.device.stage_config.assert_called_with(expected_cmds, 'cli_config')
        self.assertEqual(result, self.device.stage_config.return_value)






if __name__ == '__main__':
    unittest.main()


"""Ping another device from HPCOM7 devices.
"""
from lxml import etree
from pyhpecw7.features.errors import InvalidIPAddress
from pyhpecw7.utils.validate import valid_ip_network
from pyhpecw7.utils.xml.lib import *


class Ping(object):
    """Ping another device from an HPCOM7 device.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        host (str): IP address or name to ping from the switch
        vrf (vrf): source VRF on the switch the ping will come from
        v6 (bool): set to true if dest is v6 target
        detail (bool): set to true if you want to see per ping
            (ICMP echo request) response details

    Note:
        If an IPv6 address is provided for ``host``, there is no need
        to set ``v6`` to ``True``, but it doesn't hurt either way.

        If a name is used for ``host`` and that resolves to a v6 address,
        then ``v6`` must be set to ``True``.

    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        host (str): IP address or name to ping from the switch
        vrf (vrf): source VRF on the switch the ping will come from
        v6 (bool): set to true if dest is v6 target
        detail (bool): set to true if you want to see per ping
            (ICMP echo request) response details
        response (dict): see ``_ping``

    """

    def __init__(self, device, host, vrf='', v6=False, detail=False):
        self.device = device
        self.host = host
        self.vrf = vrf or ''
        self.v6 = v6
        self.detail = detail

        # list of XML tags used to build proper objects
        self.v4tags = ['Ping', 'IPv4Ping', 'PingTest']
        self.v6tags = ['Ping', 'IPv6Ping', 'PingTest']

        self.response = self._ping()

    def _ping(self):
        """Builds XML object for VLAN configuration and sends to staging

        Returns:
            It returns a dictionary with the following k/v pairs:

                :payload_length (str): bytes of payload
                :max (str): represents max response time in milliseconds
                :min (str): represents min response time in milliseconds
                :avg (str): represents avg response time in milliseconds
                :packets_tx (str): number of packets sent
                :packets_rx (str): number of packets received
                :host (str): target host being "pinged"
                :loss_rate (str): percent of which pings are dropped
                :detailed_response (list): list of dicts that provides
                    detail for each icmp request including icmp seq #
                    and reply time.

        """
        if '.' in self.host or ':' in self.host:
            self.param_check(host=self.host)

        E = action_element_maker()
        if ':' in self.host or self.v6 is True:
            top = E.top(
                E.Ping(
                    E.IPv6Ping(
                        E.PingTest(
                            E.Host(self.host),
                            E.VRF(self.vrf)
                        )
                    )
                )
            )
        else:
            top = E.top(
                E.Ping(
                    E.IPv4Ping(
                        E.PingTest(
                            E.Host(self.host),
                            E.VRF(self.vrf)
                        )
                    )
                )
            )

        rsp = self.device.action(top)
        return self._build_response(rsp)

    def _build_response(self, response):
        """Builds dictionary from XML response coming from device
        """
        as_string = response.xml
        as_xml = etree.fromstring(as_string.encode('ascii'))

        key_map = {
            'payload_length': 'PayloadLength',
            'packets_tx': 'TotalTransmitPacket',
            'packets_rx': 'TotalReceivePacket',
            'loss_rate': 'LossRate',
            'min': 'MinReplyTime',
            'max': 'MaxReplyTime',
            'avg': 'AvgReplyTime',
            'host': 'Host',
            'source': 'SrcAddr'
        }

        def _get_time(time):
            return str(int(find_in_action('ReplyTime', time).text) // 1000)

        ping_response = {}
        for new_key, xml_tag in key_map.items():
            value = ''
            if new_key in ['min', 'avg', 'max']:
                value = _get_time(as_xml)
            else:
                get_obj = find_in_action(xml_tag, as_xml)
                if get_obj is not None:
                    value = get_obj.text
            if value:
                ping_response[new_key] = value
            if self.detail:
                five_replies = findall_in_action('EchoReply', as_xml)
                replies = []
                for reply in five_replies:
                    icmp_seq = find_in_action('IcmpSequence', reply).text
                    reply_time = _get_time(reply)
                    temp = dict(icmp_seq=icmp_seq, reply_time=reply_time)
                    replies.append(temp)
                ping_response['detailed_response'] = replies

        return ping_response

    def param_check(self, **kvargs):
        """Basic param validation for v4 and v6 addresses
        """
        host = kvargs.get('host')
        if host:
            if not valid_ip_network(host):
                raise InvalidIPAddress(host)



"""This module is used to configure IRF for a
``HPCOM7`` device. The ``IRFMember`` class is used
to change the IRF member numbers on the device, and
can force a reboot. The ``IRFPort`` class is used to
bind physical ports to IRF ports. ``IRFMember`` should
be used first and the ``IRFPort``.
"""

from pyhpecw7.utils.xml.lib import *
from pyhpecw7.features.interface import Interface
from pyhpecw7.features.errors import InterfaceAbsentError,\
    IRFMemberDoesntExistError


class IrfPort(object):
    """This class is used to get and build IRF port configurations
    on a ``HPCOM7`` device.

    Args:
        device (HPCOM7): connected instance of
            a ``pyhpecw7.comware.HPCOM7`` object.

    Attributes:
        device (HPCOM7): connected instance of
            a ``pyhpecw7.comware.HPCOM7`` object.
    """
    def __init__(self, device):
        self.device = device

    def _build_iface_updown(self, iface_list, updown):
        """Stage commands to bring the interfaces up or down.
        """
        for iface_name in iface_list:
            iface = Interface(self.device, iface_name)
            if not iface.iface_exists:
                raise InterfaceAbsentError(iface.interface_name)
            iface.build(stage=True, admin=updown)

    def build(self, member_id, old_p1=[], old_p2=[],
              irf_p1=[], irf_p2=[],
              filename='startup.cfg',
              activate=True):
        """Stage all of the commands to configure IRF ports, including:
        1. Bringing down interfaces to be removed or added
        2. Binding physical ports to IRF ports
        3. Bringing up interfaces to be added.
        4. Saving the config.
        5. (Optionally) activating the IRF port configuration.

        Args:
            member_id (string): The member ID of the switch.
            old_p1 (list): REQUIRED - A list of current interfaces
                bound to IRF port 1.
            old_p2 (list): REQUIRED - A list of current interfaces
                bound to IRF port 2.
            irf_p1 (list): REQUIRED - A list of desired interfaces to
                be bound to IRF port 1.
            irf_p2 (list): REQUIRED - A list of desired interfaeces to
                be bound to IRF port 2.
            filename (str): OPTIONAL - The filename in which to
                save the current configuration.
                Defaults to 'startup.cfg'.
            activate (bool): OPTIONAL - Whether to immediately
                apply the IRF port configuration.
                Defaults to ``True``.

        Note:
            The ``irf_p1`` and ``irf_p2`` should be the complete physical
            interface list for IRF port 1 and IRF port 2, respectively.
            Interfaces not in the list will be removed.

        Note:
            If ``old_p1`` and ``old_p2`` are not accurate, behavior is
            undefined. These values can be obtained from ``get_config()``.

        Returns:
            A string representation of the list of
            staged configurations on the device.
        """
        updown_ifaces = []
        updown_ifaces += list(set(old_p1).difference(irf_p1))
        updown_ifaces += list(set(irf_p1).difference(old_p1))
        updown_ifaces += list(set(old_p2).difference(irf_p2))
        updown_ifaces += list(set(irf_p2).difference(old_p2))

        self._build_iface_updown(set(updown_ifaces), 'down')

        EN = nc_element_maker()
        EC = config_element_maker()

        operation1 = None
        if not irf_p1:
            operation1 = 'remove'

        iface1_eles = []
        for iface_name in irf_p1:
            iface1_eles.append(EC.Interface(EC.IfName(iface_name)))

        operation2 = None
        if not irf_p2:
            operation2 = 'remove'

        iface2_eles = []
        for iface_name in irf_p2:
            iface2_eles.append(EC.Interface(EC.IfName(iface_name)))

        config = EN.config(
            EC.top(
                EC.IRF(
                    EC.IRFPorts(
                        EC.IRFPort(
                            EC.MemberID(member_id),
                            EC.Port('1'),
                            *iface1_eles,
                            **operation_kwarg(operation1)
                        ),
                        EC.IRFPort(
                            EC.MemberID(member_id),
                            EC.Port('2'),
                            *iface2_eles,
                            **operation_kwarg(operation2)
                        ),
                    )
                )
            )
        )

        self.device.stage_config(config, 'edit_config')
        self._build_iface_updown(set(irf_p1 + irf_p2), 'up')

        self.device.stage_config(filename, 'save')

        if activate:
            EA = action_element_maker()
            top = EA.top(
                EA.IRF(
                    EA.PortConfiguration(
                        EA.Activate()
                    )
                )
            )

            self.device.stage_config(top, 'action')

        return self.device.staged_to_string()

    def get_config(self):
        """Get the current configuration of IRF ports on the device.

        Returns:
            A dictionary of IRF port bindings.

            It has the following format::

                {
                    <member_id>: {
                        'irf_p1' : <iface_list>,
                        'irf_p2' : <iface_list>
                    }
                }
        """
        irf_ports = {}

        E = data_element_maker()
        top = E.top(
            E.IRF(
                E.IRFPorts(
                    E.IRFPort(
                        E.Interface()
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        port_elems = findall_in_data('IRFPort', nc_get_reply.data_ele)

        for port_ele in port_elems:
            iface_elems = findall_in_data('Interface', port_ele)
            port_num = find_in_data('Port', port_ele).text
            member_num = find_in_data('MemberID', port_ele).text

            if irf_ports.get(member_num) is None:
                irf_ports[member_num] = {}

            member_dict = irf_ports[member_num]
            if len(iface_elems):
                for iface_ele in iface_elems:
                    iface_name = find_in_data('IfName', iface_ele).text

                    if port_num == '1':
                        if member_dict.get('irf_p1') is None:
                            member_dict['irf_p1'] = []
                        irf_ports[member_num]['irf_p1'].append(iface_name)
                    elif port_num == '2':
                        if member_dict.get('irf_p2') is None:
                            member_dict['irf_p2'] = []
                        irf_ports[member_num]['irf_p2'].append(iface_name)

        return irf_ports


class IrfMember(object):
    """This class is used to get and build IRF port configurations
    on a ``HPCOM7`` device.

    Args:
        device (HPCOM7): connected instance of
            a ``pyhpecw7.comware.HPCOM7`` object.

    Attributes:
        device (HPCOM7): connected instance of
            a ``pyhpecw7.comware.HPCOM7`` object.
    """
    def __init__(self, device):
        self.device = device

    def _get_member_config(self, member_id):
        """Get membership configuration for the specified member ID.
        Information includes priority, description, and the new member ID.
        """
        irf_members = {}
        key_map = {'new_member_id': 'NewMemberID',
                   'descr': 'Description',
                   'priority': 'Priority'}

        E = data_element_maker()
        top = E.top(
            E.IRF(
                E.Members(
                    E.Member()
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        member_eles = findall_in_data('Member', nc_get_reply.data_ele)

        for member_ele in member_eles:
            member_num = find_in_data('MemberID', member_ele).text
            irf_members[member_num] = data_elem_to_dict(member_ele, key_map)

        return irf_members.get(member_id)

    def _build_member_config(self, **params):
        """Build the IRF membership configuration.
        Configuration parameters include priority,
        description, and the new member id.

        Possible parameters:
            member_id (str)
            new_member_id (str)
            priority (str)
            descr (str)

        Returns:
            True is successful.
        """
        member_id = params.get('member_id')
        descr = params.get('descr')
        if descr:
            if not self.device.stage_config(
                    'irf member {0} description {1}'.format(
                        member_id, descr), 'cli_config'):
                return False

        new_member_id = params.get('new_member_id')
        if new_member_id:
            if not self.device.stage_config(
                    'irf member {0} renumber {1}'.format(
                        member_id, new_member_id), 'cli_config'):
                return False

        priority = params.get('priority')
        if priority:
            if not self.device.stage_config(
                    'irf member {0} priority {1}'.format(
                        member_id, priority), 'cli_config'):
                return False

        return True

    def _get_domain_config(self):
        """Return the domain id of switch or IRF stack.

        Returns:
            The domain id for the switch or IRF stack.
        """
        key_map = {'domain_id': 'Domain'}

        E = data_element_maker()
        top = E.top(
            E.IRF(
                E.Configuration()
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        reply_data = find_in_data('Configuration', nc_get_reply.data_ele)

        domain_data = data_elem_to_dict(reply_data, key_map)

        return domain_data

    def _get_auto_update_config(self):
        """Return whether the IRF auto update feature is enabled.

        Returns:
            'enable' or 'disable'
        """
        key_map = {'auto_update': 'AutoUpgrade'}

        E = data_element_maker()
        top = E.top(
            E.IRF(
                E.Configuration()
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        reply_data = find_in_data('Configuration', nc_get_reply.data_ele)

        au_data = data_elem_to_dict(reply_data, key_map)

        return au_data

    def _build_auto_update_domain_config(self, auto_update, domain_id):
        """Build the auto update config iformation.

        Args:
            auto_update (str): 'enable' or 'disable'
            domain_id (str): The domain id for the switch

        Returns:
            The staged config.
        """
        EN = nc_element_maker()
        EC = config_element_maker()

        data_eles = []
        if auto_update:
            data_eles.append(EC.AutoUpgrade(auto_update))

        if domain_id:
            data_eles.append(EC.Domain(domain_id))

        config = EN.config(
            EC.top(
                EC.IRF(
                    EC.Configuration(*data_eles)
                )
            )
        )

        return self.device.stage_config(config, 'edit_config')

    def _get_mad_exclude(self):
        """Get the currently configured mad excluded interfaces

        Returns:
            A dictionary::
                {
                    'mad_exclude': <interface list>
                }
        """
        mad_ex_ifaces = []
        raw_rsp = self.device.cli_display(
            'display current | inc "mad exclude interface"')
        lines = raw_rsp.split('\n')[1:-1]

        for line in lines:
            mad_ex_ifaces.append(line.split()[-1])

        return {'mad_exclude': mad_ex_ifaces}

    def _build_mad_exclude(self, iface_list):
        """Build the configuration for adding interfaces
        to the mad exclude list.
        """
        for iface_name in iface_list:
            iface = Interface(self.device, iface_name)
            if not iface.iface_exists:
                raise InterfaceAbsentError(iface.interface_name)
            if not self.device.stage_config(
                    'mad exclude interface {0}'.format(
                        iface.interface_name), 'cli_config'):
                return False
        return True

    def get_config(self, member_id):
        """Return the entire IRF membership configuration
        for a given member ID.

        Args:
            member_id (str): A current IRF member ID for
                the device.

        Returns:
            A dictionary with IRF membership configuration parameters.
            Including:
                :auto_update (str): 'enable' or 'disable'.
                :mad_exclude (list): A list of interfaces
                    to be excluded from MAD mechanisms.
                :new_member_id (str): The new IRF member
                    ID of the device to applied after reboot.
                :descr (str): A textual description of the IRF member.
                :priority (str): The priority of the IRF member.
                :domain_id (str): The domain id of the member.

        Raises:
            IRFMemberDoesntExistError: if the IRF member doesn't exist.
        """
        member_config = self._get_member_config(member_id)
        if member_config is None:
            raise IRFMemberDoesntExistError(member_id)
        au_config = self._get_auto_update_config()
        madex_config = self._get_mad_exclude()
        domain_config = self._get_domain_config()

        member_config.update(au_config)
        member_config.update(madex_config)
        member_config.update(domain_config)

        return member_config

    def build(self, **params):
        """Build the IRF membership configuration.

        Args:
            **params: See Keyword Args.

        Keyword Args:
            auto_update (str): OPTIONAL - Enables or disables
                the auto-upgrade of software after an IRF
                convergence. Should be 'enable' or 'disable'.
            mad_exclude (list): OPTIONAL - A list of interfaces
                to be excluded from MAD mechanisms.
            member_id (str): REQUIRED - The current IRF member ID
                of the device.
            new_member_id (str) REQUIRED - The desired new IRF member
                ID of the device.
            domain_id (str) OPTIONAL - The domain id for the IRF member
            descr (str): OPTIONAL - A textual description of the IRF member.
            priority (str): OPTIONAL - The desired priority of the IRF member.

        Returns:
            True if successful, False otherwise.
        """
        if not self._build_member_config(**params):
            return False

        auto_update = params.get('auto_update')
        domain_id = params.get('domain_id')
        if auto_update or domain_id:
            if not self._build_auto_update_domain_config(auto_update, domain_id):
                return False

        mad_exclude = params.get('mad_exclude')
        if mad_exclude:
            if not self._build_mad_exclude(mad_exclude):
                return False

        return True

    def remove_mad_exclude(self, iface_list):
        """Stage the configuration to remove interfaces
        from the mad excluded list.
        """
        for iface_name in iface_list:
            self.device.stage_config(
                'undo mad exclude interface {0}'.format(
                    iface_name), 'cli_config')



import unittest
import mock
from tempfile import NamedTemporaryFile

try:
    import __builtin__
except ImportError:
    import builtins as __builtin__

from pyhpecw7.features.file_copy import FileCopy, FileNotEnoughSpaceError, FileNotReadableError, FileHashMismatchError, FileTransferError, NCError
from .base_feature_test import BaseFeatureCase

SOURCE_FILE = '/path/to/source/file.txt'

class FileCopyTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self, mock_device):
        self.device = mock_device
        self.file_copy = FileCopy(self.device, SOURCE_FILE)

    def test_init(self):
        self.assertEqual(self.file_copy._remote_dir, 'flash:/')
        self.assertEqual(self.file_copy.src, SOURCE_FILE)
        self.assertEqual(self.file_copy.dst, 'flash:/file.txt')
        self.assertEqual(self.file_copy.port, 22)
        self.assertEqual(self.file_copy.remote_dir_exists, True)

    def test_get_flash_size(self):
        self.device.cli_display.return_value = self.read_cli_display('dir_flash')
        result = self.file_copy._get_flash_size()

        self.assertEqual(result, 742204000)

        self.device.cli_display.return_value = 'garbaldigook'
        result = self.file_copy._get_flash_size()

        self.assertEqual(result, 0)

    @mock.patch.object(FileCopy, '_get_flash_size')
    @mock.patch('os.path.getsize')
    def test_enough_space(self, mock_getsize, mock_get_flash_size):
        mock_getsize.return_value = 2
        mock_get_flash_size.return_value = 1

        with self.assertRaises(FileNotEnoughSpaceError):
            self.file_copy._enough_space()

        mock_getsize.return_value = 1
        mock_get_flash_size.return_value = 2

        result = self.file_copy._enough_space()
        self.assertEqual(result, None)

    @mock.patch.object(FileCopy, '_get_remote_md5')
    @mock.patch.object(FileCopy, '_get_local_md5')
    def test_file_already_exists(self, mock_local_md5, mock_remote_md5):
        mock_local_md5.return_value = 'abc123'
        mock_remote_md5.return_value = 'abc123'

        result = self.file_copy.file_already_exists()
        self.assertEqual(result, True)

        self.file_copy.remote_dir_exists = False
        result = self.file_copy.file_already_exists()
        self.assertEqual(result, False)

        self.file_copy.remote_dir_exists = True
        mock_remote_md5.return_value = None
        result = self.file_copy.file_already_exists()
        self.assertEqual(result, False)

        mock_remote_md5.side_effect = NCError
        result = self.file_copy.file_already_exists()
        self.assertEqual(result, False)

    @mock.patch.object(__builtin__, 'open')
    def test_safety_checks_file_not_readable(self, mock_open):
        mock_open.side_effect = IOError
        with self.assertRaises(FileNotReadableError):
            self.file_copy._safety_checks()

    @mock.patch.object(__builtin__, 'open')
    def test_safety_checks_no_remote_dir(self, mock_open):
        self.file_copy.remote_dir_exists = False
        with self.assertRaises(FileRemoteDirDoesNotExist):
            self.file_copy._safety_checks()

    @mock.patch.object(__builtin__, 'open')
    def test_safety_checks_no_remote_dir(self, mock_open):
        self.file_copy.remote_dir_exists = False
        with self.assertRaises(FileRemoteDirDoesNotExist):
            self.file_copy._safety_checks()

    @mock.patch.object(FileCopy, 'file_already_exists')
    @mock.patch.object(FileCopy, '_enough_space')
    @mock.patch.object(__builtin__, 'open')
    def test_safety_checks_no_remote_dir(self, mock_open, mock_enough_space, mock_file_already_exists):
        mock_file_already_exists.return_value = False
        self.file_copy._safety_checks()
        mock_enough_space.assert_called_with()

    def test_remote_md5(self):
        expected_get, get_reply = self.xml_action_and_reply('file_copy_remote_md5')
        self.device.action.return_value = get_reply

        expected = '44d5527772e1b9841f99cb03f31cbc1c'
        result = self.file_copy._get_remote_md5()

        self.assertEqual(result, expected)
        self.assert_action_request(expected_get)

    def test_local_md5(self):
        test_file = NamedTemporaryFile()
        self.file_copy.src = test_file.name

        test_file.write('Test content.'.encode('ascii'))
        test_file.flush()

        result = self.file_copy._get_local_md5()
        self.assertEqual(result, 'bcb898f62d9e1ac765c77e6804cbd872')

        test_file.close()

    def test_remote_dir(self):
        expected_get, get_reply = self.xml_get_and_reply('file_copy_remote_dir')
        self.device.get.return_value = get_reply

        expected = True
        result = self.file_copy._remote_dir_exists()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)


    @mock.patch('pyhpecw7.features.file_copy.paramiko')
    @mock.patch('pyhpecw7.features.file_copy.SCPClient')
    @mock.patch.object(FileCopy, '_safety_checks')
    @mock.patch.object(FileCopy, '_get_local_md5')
    @mock.patch.object(FileCopy, '_get_remote_md5')
    def test_transfer_file(self, mock_remote_md5, mock_local_md5, mock_safety_checks, mock_SCP, mock_paramiko):
        mock_remote_md5.return_value = 'abc'
        mock_local_md5.return_value = 'abc'

        mock_ssh = mock_paramiko.SSHClient.return_value

        self.file_copy.transfer_file()

        mock_paramiko.SSHClient.assert_called_with()

        mock_ssh.set_missing_host_key_policy.assert_called_with(mock_paramiko.AutoAddPolicy.return_value)
        mock_ssh.connect.assert_called_with(allow_agent=False,
                                             hostname=self.device.host,
                                             look_for_keys=False,
                                             password=self.device.password,
                                             port=22,
                                             username=self.device.username)

        mock_SCP.assert_called_with(mock_ssh.get_transport.return_value)
        mock_SCP.return_value.put.assert_called_with('/path/to/source/file.txt', 'flash:/file.txt')
        mock_SCP.return_value.close.assert_called_with()

    @mock.patch('pyhpecw7.features.file_copy.paramiko')
    @mock.patch('pyhpecw7.features.file_copy.SCPClient')
    @mock.patch.object(FileCopy, '_safety_checks')
    @mock.patch.object(FileCopy, '_get_local_md5')
    @mock.patch.object(FileCopy, '_get_remote_md5')
    def test_transfer_file_mismatch_hash(self, mock_remote_md5, mock_local_md5, mock_safety_checks, mock_SCP, mock_paramiko):
        mock_remote_md5.return_value = 'abc'
        mock_local_md5.return_value = 'def'

        mock_ssh = mock_paramiko.SSHClient.return_value

        with self.assertRaises(FileHashMismatchError):
            self.file_copy.transfer_file()

        mock_paramiko.SSHClient.assert_called_with()

        mock_ssh.set_missing_host_key_policy.assert_called_with(mock_paramiko.AutoAddPolicy.return_value)
        mock_ssh.connect.assert_called_with(allow_agent=False,
                                             hostname=self.device.host,
                                             look_for_keys=False,
                                             password=self.device.password,
                                             port=22,
                                             username=self.device.username)

        mock_SCP.assert_called_with(mock_ssh.get_transport.return_value)
        mock_SCP.return_value.put.assert_called_with('/path/to/source/file.txt', 'flash:/file.txt')
        mock_SCP.return_value.close.assert_called_with()

    @mock.patch('pyhpecw7.features.file_copy.paramiko')
    @mock.patch('pyhpecw7.features.file_copy.SCPClient')
    @mock.patch.object(FileCopy, '_safety_checks')
    @mock.patch.object(FileCopy, '_get_local_md5')
    @mock.patch.object(FileCopy, '_get_remote_md5')
    def test_transfer_file_error(self, mock_remote_md5, mock_local_md5, mock_safety_checks, mock_SCP, mock_paramiko):
        mock_remote_md5.return_value = 'abc'
        mock_local_md5.return_value = 'def'

        mock_ssh = mock_paramiko.SSHClient.return_value

        mock_SCP.return_value.put.side_effect = Exception

        with self.assertRaises(FileTransferError):
            self.file_copy.transfer_file()

        mock_paramiko.SSHClient.assert_called_with()

        mock_ssh.set_missing_host_key_policy.assert_called_with(mock_paramiko.AutoAddPolicy.return_value)
        mock_ssh.connect.assert_called_with(allow_agent=False,
                                             hostname=self.device.host,
                                             look_for_keys=False,
                                             password=self.device.password,
                                             port=22,
                                             username=self.device.username)

        mock_SCP.assert_called_with(mock_ssh.get_transport.return_value)
        mock_SCP.return_value.put.assert_called_with('/path/to/source/file.txt', 'flash:/file.txt')


    def test_create_dir(self):
        self.file_copy._remote_dir = 'flash:/unit/'

        expected_get, get_reply = self.xml_action_and_reply('file_copy_create_remote_dir')
        self.device.action.return_value = get_reply

        expected = None
        result = self.file_copy.create_remote_dir()

        self.assertEqual(result, expected)
        self.assert_action_request(expected_get)


if __name__ == "__main__":
    unittest.main()



"""Manage layer 3 interfaces on HPCOM7 devices.
"""
from pyhpecw7.utils.network import ipaddr
from pyhpecw7.features.interface import Interface
from pyhpecw7.features.errors import IpIfaceMissingData
from pyhpecw7.utils.xml.lib import *

V4 = 'v4'
V6 = 'v6'


class IpInterface(object):
    """This class is used to get and build layer 3
    interface configurations on ``HPCOM7`` devices.

    Args:
        device (HPCOM7): connected instance of
            a ``pyhpecw7.comware.HPCOM7`` object.
        interface_name (str): The name of the interface.
        version (str): V4 for IPv4, V6 for IPv6

    Attributes:
        device (HPCOM7): connected instance of
            a ``pyhpecw7.comware.HPCOM7`` object.
        interface_name (str): The name of the interface.
        version (str): V4 for IPv4, V6 for IPv6
        interface (pyhpecw7.features.Interface): The associated
            ``Interface`` configuration object.
        is_routed (bool): ``True`` if the interface is in layer 3
            mode, ``False`` otherwise.
    """
    def __init__(self, device, interface_name, version=V4):
        self.device = device
        self.interface = Interface(device, interface_name)
        self.interface_name = self.interface.interface_name
        self.is_routed = self.interface.is_routed
        self.version = version

    def gen_ipv4_top(self):
        E = data_element_maker()
        top = E.top(
            E.IPV4ADDRESS(
                E.Ipv4Addresses(
                    E.Ipv4Address(
                        E.IfIndex(self.interface.iface_index),
                        E.Ipv4Address(),
                        E.Ipv4Mask(),
                    )
                )
            )
        )

        return top

    def gen_ipv6_top(self):
        E = data_element_maker()
        top = E.top(
            E.IPV6ADDRESS(
                E.Ipv6Addresses(
                    E.AddressEntry(
                        E.IfIndex(self.interface.iface_index),
                        E.Ipv6Address(),
                        E.Ipv6PrefixLength(),
                    )
                )
            )
        )

        return top

    def gen_ipv4_config(self, params, key_map, operation):
        EN = nc_element_maker()
        EC = config_element_maker()

        config = EN.config(
            EC.top(
                EC.IPV4ADDRESS(
                    EC.Ipv4Addresses(
                        EC.Ipv4Address(
                            EC.IfIndex(self.interface.iface_index),
                            *config_params(params, key_map)
                        )
                    ),
                    **operation_kwarg(operation)
                )
            )
        )

        return config

    def gen_ipv6_config(self, params, key_map, operation):
        EN = nc_element_maker()
        EC = config_element_maker()

        config = EN.config(
            EC.top(
                EC.IPV6ADDRESS(
                    EC.Ipv6AddressesConfig(
                        EC.AddressEntry(
                            EC.IfIndex(self.interface.iface_index),
                            *config_params(params, key_map)
                        )
                    ),
                    **operation_kwarg(operation)
                )
            )
        )

        return config

    def get_config(self):
        """Return a list of currently configured IP addresses on the interface.

        Note:
            Either only IPv4 or only IPv6 addresses will be returned depending
            on the version stored in ``self.version``.

        Returns:
            A list of currently configured IP addresses on the interface.
        """
        if self.version == V4:
            top = self.gen_ipv4_top()
            addr_tag = 'Ipv4Address'
            mask_tag = 'Ipv4Mask'
            search_tag = 'Ipv4Address'
        elif self.version == V6:
            top = self.gen_ipv6_top()
            addr_tag = 'Ipv6Address'
            mask_tag = 'Ipv6PrefixLength'
            search_tag = 'AddressEntry'
        else:
            return {}

        key_map = {'addr': addr_tag,
                   'mask': mask_tag}

        nc_get_reply = self.device.get(('subtree', top))
        reply_data = findall_in_data(search_tag, nc_get_reply.data_ele)

        existing_list = []
        for match in reply_data:
            to_dict = data_elem_to_dict(match, key_map)
            if to_dict:
                existing_list.append(to_dict)

        return existing_list

    def build(self, stage=False, **params):
        """Stage or execute a configuration to configure
        an IP address on an interface.

        Args:
            stage (bool): whether to stage the commands or execute
                immediately
            **params: see Keyword Args

        Keyword Args:
            addr (str): The IP address to add.
            mask (str): The network mask for the IP address,
                in dotted decimal or prefix length notation.

        Returns:
            True if stage=True and staging is successful
            etree.Element XML response if immediate execution
        """
        config = self._build_config(state='present', **params)
        if stage:
            return self.device.stage_config(config, 'edit_config')
        else:
            return self.device.edit_config(config)

    def remove(self, stage=False, **params):
        """Stage or execute a configuration to remove
        an IP address on an interface.

        Args:
            stage (bool): whether to stage the commands or execute
                immediately
            **params: see Keyword Args

        Keyword Args:
            addr (str): The IP address to remove.
            mask (str): The network mask for the IP address,
                in dotted decimal or prefix length notation.

        Returns:
            True if stage=True and staging is successful
            etree.Element XML response if immediate execution
        """
        config = self._build_config(state='absent', **params)
        if stage:
            return self.device.stage_config(config, 'edit_config')
        else:
            return self.device.edit_config(config)

    def _build_config(self, state='present', **params):
        """Build an etree.Element object to configure or remove
        an IP address on an interface.

        Args:
            state (str): 'present' to add an IP address,
                'absent' to remove an IP address.
            **params: see Keyword Args

        Keyword Args:
            addr (str): The IP address to add or remove.
            mask (str): The network mask for the IP address,
                in dotted decimal or prefix length notation.

        Returns:
            An etree.Element object to configure or remove
            an IP address on an interface.
        """
        address = params.get('addr')
        mask = params.get('mask')

        if not address or not mask:
            raise IpIfaceMissingData

        ip_obj = ipaddr.IPNetwork(address + '/' + mask)

        if self.version == V4:
            addr_tag = 'Ipv4Address'
            mask_tag = 'Ipv4Mask'
            params['mask'] = str(ip_obj.netmask)
        elif self.version == V6:
            addr_tag = 'Ipv6Address'
            mask_tag = 'Ipv6PrefixLength'
            params['mask'] = str(ip_obj.prefixlen)
            params['addr_origin'] = '1'
        else:
            return

        key_map = {'addr': addr_tag,
                   'mask': mask_tag,
                   'addr_origin': 'AddressOrigin'}

        if state == 'present':
            operation = 'merge'
        elif state == 'absent':
            operation = 'remove'
            del params['mask']

        if self.version == V4:
            config = self.gen_ipv4_config(params, key_map, operation)
        elif self.version == V6:
            config = self.gen_ipv6_config(params, key_map, operation)

        return config



"""Feature-specific errors.
"""
from pyhpecw7.errors import PYHPError


class FeatureError(PYHPError):
    def __init__(self):
        pass


class LengthOfStringError(FeatureError):

    def __init__(self, param_name):
        # passing in the name of the variable rather than the actual string
        # but feel free to pass in the value instead if you want!
        self.param_name = param_name

    def __repr__(self):
        errstr = 'Maximum string length of exceeded for {0}'.format(
            self.param_name)
        return errstr

    __str__ = __repr__


class InvalidIPAddress(FeatureError):

    def __init__(self, ipaddr):
        self.ipaddr = ipaddr

    def __repr__(self):
        errstr = 'Invalid IPv4 or IPv6 Address: {0}'.format(
            self.ipaddr)
        return errstr

    __str__ = __repr__


##################################
#       INTERFACE ERRORS         #
##################################


class InterfaceError(FeatureError):
    def __init__(self, if_name):
        self.if_name = if_name


class InterfaceTypeError(InterfaceError):
    def __init__(self, if_name, if_types=None):
        super(InterfaceTypeError, self).__init__(if_name)
        self.if_types = if_types

    def __repr__(self):
        errstr = '{0} is not a valid interface type.'.format(self.if_name)
        if self.if_types:
            errstr += ' Type must be one of {0}'.format(self.if_types)

        return errstr

    __str__ = __repr__


class InterfaceAbsentError(InterfaceError):
    def __init__(self, if_name):
        super(InterfaceAbsentError, self).__init__(if_name)

    def __repr__(self):
        return 'Interface {0} not found on the device.'.format(self.if_name)

    __str__ = __repr__


class InterfaceParamsError(InterfaceError):
    def __init__(self, if_name, params):
        super(InterfaceParamsError, self).__init__(if_name)
        self.params = params

    def __repr__(self):
        return 'Interface {0} does not take parameters {1}.'.format(
            self.if_name, self.params)

    __str__ = __repr__


class InterfaceCreateError(InterfaceError):
    def __init__(self, if_name):
        super(InterfaceCreateError, self).__init__(if_name)

    def __repr__(self):
        return 'Interface {0} could not be created.'.format(self.if_name)

    __str__ = __repr__


class InterfaceRemoveError(InterfaceError):
    def __init__(self, if_name):
        super(InterfaceRemoveError, self).__init__(if_name)

    def __repr__(self):
        return 'Interface {0} could not be removed.'.format(self.if_name)

    __str__ = __repr__


class InterfaceVlanMustExist(InterfaceError):
    def __init__(self, if_name, number):
        super(InterfaceVlanMustExist, self).__init__(if_name)
        self.number = number

    def __repr__(self):
        return 'Vlan {0} must exist before interface can be created.'.format(
            self.number)

    __str__ = __repr__


######################
# IPINTERFACE ERRORS #
######################

class IpInterfaceError(FeatureError):
    pass

class IpIfaceMissingData(IpInterfaceError):
    def __init__(self):
        super(IpIfaceMissingData, self).__init__()

    def __repr__(self):
        return 'IP address and mask must be supplied'

    __str__ = __repr__


##################################
#       VLAN ERRORS             #
##################################


class VlanError(FeatureError):
    pass


class VlanIDError(VlanError):

    def __repr__(self):
        errstr = 'VLAN ID must be between 1-4094'

        return errstr

    __str__ = __repr__

##################################
#       REBOOT ERRORS            #
##################################


class RebootError(FeatureError):
    pass


class RebootTimeError(RebootError):

    def __repr__(self):
        errstr = 'Format for time must be HH:MM'
        return errstr

    __str__ = __repr__


class RebootDateError(RebootError):

    def __repr__(self):
        errstr = 'Format for the date must be MM/DD/YYYY'
        return errstr

    __str__ = __repr__


##################################
#       PORTCHANNEL ERRORS       #
##################################


class PortChannelError(FeatureError):
    def __init__(self):
        pass


class InvalidPortType(PortChannelError):
    def __init__(self, if_name, config_type, pc_type):
        self.if_name = if_name
        self.config_type = config_type
        self.pc_type = pc_type

    def __repr__(self):
        errstr = ('Proposed port-channel type of "{0}" '.format(self.pc_type)
                  + '\ndoes not match existing physical interface '
                  + '\nof port type "{0}" '.format(self.config_type)
                  + 'on interface: "{0}"'.format(self.if_name))
        return errstr

    __str__ = __repr__


class AggregationGroupError(PortChannelError):
    def __init__(self, if_name):
        self.if_name = if_name

    def __repr__(self):
        errstr = ('interface {0}'.format(self.if_name)
                  + ' is assigned \nto another aggregation group.'
                  + 'It needs to be \nremoved first.')

        return errstr

    __str__ = __repr__

##################################
#       FILE COPY ERRORS         #
##################################


class FileError(FeatureError):
    def __init__(self, src=None, dst=None):
        self.src = src
        self.dst = dst


class FileNotReadableError(FileError):
    def __repr__(self):
        return '{0} doesn\'t exist, or isn\'t readable.'.format(self.src)

    __str__ = __repr__


class FileNotEnoughSpaceError(FileError):
    def __init__(self, src, file_size, flash_size):
        super(FileNotEnoughSpaceError, self).__init__(src)
        self.file_size = file_size
        self.flash_size = flash_size

    def __repr__(self):
        return 'Not enough space on remote device for {0}.\n'.format(self.src) +\
            'File Size: {0} bytes\n'.format(self.file_size) +\
            'Space Available: {0} bytes\n'.format(self.flash_size)

    __str__ = __repr__


class FileTransferError(FileError):
    def __repr__(self):
        return 'There was an error while the file was in transit.'

    __str__ = __repr__


class FileHashMismatchError(FileError):
    def __init__(self, src, dst, src_hash, dst_hash):
        super(FileHashMismatchError, self).__init__(src, dst)
        self.src_hash = src_hash
        self.dst_hash = dst_hash

    def __repr__(self):
        return 'The MD5 hash digests do not match.\n' +\
            'The hash of the source {0} was {1}.\n'.format(self.src, self.src_hash) +\
            'The hash of the destinatino {0} was {1}.\n'.format(self.dst, self.dst_hash)

    __str__ = __repr__


class FileRemoteDirDoesNotExist(FileError):
    def __init__(self, remote_dir):
        self.remote_dir = remote_dir

    def __repr__(self):
        return 'The remote directory {0}'.format(self.remote_dir) +\
            ' does not exist.'

    __str__ = __repr__


##################################
#       Config Errors            #
##################################

class ConfigError(FeatureError):
    pass


class InvalidConfigFile(ConfigError):

    def __repr__(self):
        errstr = ('Config replace operation failed.\n' +
                  ' Validate the config file being applied.')

        return errstr

    __str__ = __repr__

##################################
#       IRF Errors              #
##################################


class IRFError(FeatureError):
    pass


class IRFMemberDoesntExistError(IRFError):

    def __init__(self, member_id):
        self.member_id = member_id

    def __repr__(self):
        return 'The IRF member {0}'.format(self.member_id) +\
            ' does not exist.'

    __str__ = __repr__



"""Manage connections to HPCOM7 devices.

(c) Copyright 2016 Hewlett Packard Enterprise Development LP Licensed under the Apache License, Version 2.0
(the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License
at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
express or implied. See the License for the specific language governing permissions and limitations under the License.

"""
from ncclient import manager
from ncclient.operations.rpc import RPCError
import ncclient.transport.errors as NcTransErrors
import ncclient.operations.errors as NcOpErrors
from pyhpecw7.features.facts import Facts
import time
import socket
from lxml import etree
from pyhpecw7.utils.xml.namespaces import NETCONFBASE_C
from pyhpecw7.errors import NCTimeoutError, ConnectionClosedError, NCError,\
    ConnectionAuthenticationError, ConnectionSSHError, ConnectionUnkownHostError,\
    ConnectionError, LockConflictError, UnlockConflictError


class HPCOM7(object):
    """This class manages the NETCONF connection to an HP Comware switch,
    and provides methods to execute various NETCONF operations.

    Args:
        host: REQUIRED - hostname or IP address of the HP Comware Switch
        username: REQUIRED - username used to login to the switch. Requires
            network-admin access
        password: REQUIRED - password used to login to the switch. Requires
            network-admin access
        port: OPTIONAL - integer that represents the NETCONF port being
            used on the switch.  Default is 830.
        timeout: OPTIONAL - How long a single RPC rquest should wait
            before timing out.
        ssh_config: OPTIONAL - enables parsing of a OpenSSH configuration
            file, either file in string, or defaults to ~/.ssh/config if True

    Attributes:
        staged: Dictionary that stores XML objects prior to being sent to
           the device.  It is a "holding" area.  The stage_config method
           is are used to build this attribute.  Keys are user defined
           to make it possible to recall/view the specified object if more
           than one are being prepared to send.
        staged_cli: Dictionary that stores XML objects when pushing
           raw CLI commands via NETCONF.  It is a "holding" area.
           The stage_cli_display and stage_cli_config methods are used
           to build this attribute.  Keys are user defined to make it
           possible to recall/view the specified object if more than one
           are being prepared to send.
    """
    def __init__(self, **kvargs):
        self.host = kvargs.get('host')
        self.username = kvargs.get('username')
        self.password = kvargs.get('password')
        self.port = kvargs.get('port') or 830
        self.timeout = kvargs.get('timeout') or 30
        self.ssh_config = kvargs.get('ssh_config', None)
        self.staged = []

        self._locked = False

    def open(self,
             hostkey_verify=False,
             allow_agent=True,
             look_for_keys=False):
        """Open the NETCONF connection to the HP switch.

        Args:
            hostkey_verify (bool): OPTIONAL - enables hostkey verification
                from ~/.ssh/known_hosts
            allow_agent (bool): OPTIONAL -
                enables querying SSH agent (if found) for keys
            look_for_keys (bool): OPTIONAL - enables looking in the
                usual locations for ssh keys (e.g. ~/.ssh/id_*)

        Returns:
            Connection to the device using ncclient's connect method.

        Raises:
            ConnectionAuthenticationError: if there is an error authenticating
                to the device.
            ConnectionSSHError: if NETCONF isn't enabled on the device, or the
                device isn't reachable
            ConnectionUnkownHostError: if the device's network name cannot
                be resolved to an IP address.
            ConnectionError: if an unkown error occurs during connection
        """
        time.sleep(.25)

        try:
            self.connection = manager.connect(host=self.host,
                                              port=self.port,
                                              username=self.username,
                                              password=self.password,
                                              device_params={'name': 'hpcomware'},
                                              hostkey_verify=hostkey_verify,
                                              allow_agent=allow_agent,
                                              look_for_keys=look_for_keys,
                                              timeout=self.timeout,
                                              ssh_config=self.ssh_config)

        except NcTransErrors.AuthenticationError:
            raise ConnectionAuthenticationError(self)
        except NcTransErrors.SSHError:
            raise ConnectionSSHError(
                self, msg='There was an error connecting with SSH.'
                ' The NETCONF server may be down or refused the connection.'
                ' The connection may have timed out if the server wasn\'t reachable.')
        except socket.gaierror:
            raise ConnectionUnkownHostError(self)
        except ImportError:
            raise ImportError('ncclient does not have the comware extensions')
        except Exception:
            raise ConnectionError(self, msg='There was an unknown error while trying to connect.')

        return self.connection

    @property
    def facts(self):
        """
        A dictionary of a facts about the device.
        ``None`` if not connected.
        """
        if hasattr(self, 'connection'):
            if self.connection.connected:
                facts = Facts(self)
                return dict(facts.facts)
        return None

    @property
    def connected(self):
        """``True`` if the NETCONF session to the device is open
        else it is ``False``.
        """
        if hasattr(self, 'connection'):
            return self.connection.connected

        return False

    def close(self):
        """Close the NETCONF connection to the HP switch.
        """
        try:
            if self.connected:
                self.connection.close_session()
        except NcOpErrors.TimeoutExpiredError:
            raise NCTimeoutError

    def stage_config(self, config, cfg_type):
        """Append config object to the staging area.

        Args:
            config: The config payload. Could be a partial etree.Element XML
                object or raw text if using a CLI config type
            cfg_type (string): The type of config payload.
                Permitted options: "edit_config", "action", "cli_config",
                "cli_display", "save", "rollback"

        Returns:
            True if config object was successfully staged.

        Raises:
            ValueError: if an invalid config type is supplied.
        """
        if cfg_type in ['edit_config', 'action', 'cli_config',
                        'cli_display', 'save', 'rollback']:
            self.staged.append({'config': config, 'cfg_type': cfg_type})
            return True
        else:
            raise ValueError("Invalid config type for staging.  Must be one"
                             + "of the following: edit_config, action, "
                             + "cli_config, cli_display, save, rollback")

    def staged_to_string(self):
        """Convert the staging area to a list of strings.

        Returns:
            A list of string representing the configuration in the
            staging area.
        """
        cfgs = []
        for cfg in self.staged:
            if isinstance(cfg['config'], etree._Element):
                cfgs.append(etree.tostring(cfg['config']))
            else:
                cfgs.append(cfg['config'])

        return cfgs

    def execute(self, run_cmd_func, args=[], kwargs={}):
        """Safely execute the supplied function with args and kwargs.

        Args:
            run_cmd_func(executable): Function to be run.

        Returns:
            The return value of the supplied function.

        Raises:
            NCError: if there is an error in the NETCONF protocol.
            NCTimeoutError: if a client-side timeout has occured.
            ConnectionClosedError: if the NETCONF session is closed.
        """
        if self.connected is not True:
            raise ConnectionClosedError(self)

        try:
            self.lock()
            rsp = run_cmd_func(*args, **kwargs)
        except RPCError as e:
            raise NCError(e)
        except NcOpErrors.TimeoutExpiredError:
            raise NCTimeoutError
        except NcTransErrors.TransportError:
            raise ConnectionClosedError(self)
        finally:
            self.unlock()

        return rsp

    def execute_staged(self, target='running'):
        """Execute/Push the XML object(s) or CLI strings in the staging
        area (self.staged) to the device.

        Args:
            target (str): must be set to running.
                It *could* change in the future
                if HP supports candidate configurations, etc.
                Only used for 'edit_config' API calls.
                Defaults to 'running'.

        Returns:
            A list of responses received from the device.
            Responses with CLI information are extracted from the XML
            response.
        """
        rsps = []
        for command in self.staged:
            cfg_type = command['cfg_type']
            config = command['config']
            args = []
            kwargs = {}
            if cfg_type == 'edit_config':
                run_cmd_func, kwargs = self.edit_config, dict(target=target, config=config)
            elif cfg_type == 'action':
                run_cmd_func, args = self.action, [config]
            elif cfg_type == 'save':
                run_cmd_func, args = self.save, [config]
            elif cfg_type == 'rollback':
                run_cmd_func, args = self.rollback, [config]
            elif cfg_type == 'cli_config':
                run_cmd_func, args = self.cli_config, [config]
            elif cfg_type == 'cli_display':
                run_cmd_func, args = self.cli_display, [config]

            rsps.append(run_cmd_func(*args, **kwargs))

        del self.staged[:]
        return rsps

    def lock(self, target='running'):
        """Attempt to lock the NETCONF connection.

        Raises:
            NCError: if there is an error in the NETCONF protocol.
            LockConflictError: if another process hold the NETCONF lock.
        """
        try:
            self.connection.lock(target)
            self._locked = True
        except RPCError as e:
            if e.tag == 'lock-denied':
                raise LockConflictError
            else:
                raise NCError(e)

    def unlock(self, target='running'):
        """Attempt to unlock the NETCONF connection.

        Raises:
            NCError: if there is an error in the NETCONF protocol.
            LockConflictError: if another process hold the NETCONF lock.
        """
        try:
            if self._locked:
                self.connection.unlock(target)
                self._locked = False
        except RPCError as e:
            if e.tag == 'operation-failed'\
                    and 'Unlock Failed' in e.message:
                raise UnlockConflictError
            else:
                raise NCError(e)
        except NcTransErrors.TransportError:
            pass

    def edit_config(self, config, target='running'):
        """Send a NETCONF edit_config XML object to the device.

        Args:
            config: etree.Element sent to ncclient.manager.edit_config
            target: Name of configuration on the remote device. Defaults to 'running'

        Returns:
            The etree.Element returned from ncclient.manager.edit_config
        """
        rsp = self.execute(self.connection.edit_config, kwargs=dict(target=target, config=config))
        return rsp

    def get(self, get_tuple=None):
        """Wrapper for ncclient.manger.get

        Args:
            get_tuple: The tuple sent to ncclient.manager.get,
                e.g: ('subtree', <etree.Element>)

        Returns:
            The etree.Element returned from ncclient.manager.get

        """
        rsp = self.execute(self.connection.get, [get_tuple])
        return rsp

    def action(self, element):
        """Wrapper for ncclient.manger.action

        Args:
            element: etree.Element sent to ncclient.manager.action

        Returns:
            The etree.Element returned from ncclient.manager.action
        """
        rsp = self.execute(self.connection.action, [element])
        return rsp

    def save(self, filename=None):
        """Wrapper for ncclient.manger.save

        Args:
            element: etree.Element sent to ncclient.manager.save

        Returns:
            The etree.Element returned from ncclient.manager.save
        """
        rsp = self.execute(self.connection.save, [filename])
        return rsp

    def rollback(self, filename):
        """Wrapper for ncclient.manger.rollback

        Args:
            element: etree.Element sent to ncclient.manager.rollback

        Returns:
            The etree.Element returned from ncclient.manager.rollback
        """
        rsp = self.execute(self.connection.rollback, [filename])
        return rsp

    def cli_display(self, command):
        """Immediately push display commands to the device and returns text.

        Args:
            command (list or string): display commands

        Returns:
            raw text CLI output

        """
        rsp = self.execute(self.connection.cli_display, [command])
        text = self._find_between(rsp.xml, 'CDATA[', ']]')
        text = self._strip_return(text)

        return text

    def cli_config(self, command):
        """Immediately push config commands to the device and returns text.

        Args:
            command (list or string): config commands

        Returns:
            raw text CLI output

        """
        rsp = self.execute(self.connection.cli_config, [command])
        xml_obj = etree.fromstring(rsp.xml.encode('ascii'))
        return self._extract_config(xml_obj)

    def reboot(self):
        """Attempt an immediate reboot of the device.

        """
        try:
            self.connection.async_mode = True
            self.connection.cli_display(['reboot force'])
        except NCTimeoutError:
            pass
        finally:
            self.connection.async_mode = False

    def _strip_return(self, text):
        """Strip excess return characters from text.
        """
        if text:
            text = text.replace('\r\r\r', '\r')
            text = text.replace('\r\r', '\r')
            text = text.replace('\n\n\n', '\n')
            text = text.replace('\n\n', '\n')

        return text

    def _extract_config(self, xml_resp):
        """Extract a CLI response from an XML object.
        """
        conf = xml_resp.find('.//{0}Configuration'.format(NETCONFBASE_C))
        execu = xml_resp.find('.//{0}Execution'.format(NETCONFBASE_C))

        if execu is not None:
            text = execu.text
        elif conf is not None:
            text = conf.text
        else:
            text = 'Unable to extract CLI data.'

        text = self._strip_return(text)

        return text

    def _find_between(self, s, first, last):
        """Find a substring in between two other substrings.

        Args:
            s: The full string
            first: The first substring
            second: The second substring
        """
        try:
            start = s.index(first) + len(first)
            end = s.index(last, start)
            return s[start:end]
        except ValueError:
            return ""



"""Manage VRRP groups on HPCOM7 devices.
"""
from pyhpecw7.utils.templates import cli


class VRRP(object):
    """This class is used to collect data or configure a VRRP group on a
        given interface.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        interface (str): name of the Layer 3 interface
        vrid (str): virtual router ID (group number)


    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        interface (str): name of the Layer 3 interface
        vrid (str): virtual router ID (group number)


    """

    def __init__(self, device, interface, vrid):
        self.device = device
        self.interface = interface
        self.vrid = vrid

        # contains response from using CLI method
        self._existing_all = None

        # interface command used when sending CLI commands
        self.intf_command = 'interface {0}'.format(interface)

    def get_vrrp_groups(self):
        """Get list of VRIDs for a given Layer 3 interface

        Returns:
            This returns a list of VRIDs configured on a given interface.

        """
        commands = 'display vrrp interface {0} verbose'.format(self.interface)
        rsp = self.device.cli_display(commands)

        self._existing_all = cli.get_structured_data('vrrp.tmpl', rsp)

        vrids = [each.get('vrid') for each in self._existing_all
                 if each.get('vrid')]

        return vrids

    def get_auth_type(self):
        """Get auth type for a given VRID on a given interface

        Returns:
           This returns a dictionary with the following k/v pairs:
               :auth_mode (str): "simple" or "md5"
               :key_type (str): "cipher"
               :key (str): it will be a cipher

           It will return an empty dictionary if it VRID
           does not exist.

        """
        auth = {}
        commands = 'display current-configuration interface' \
                   + ' {0} | inc "vrid {1} auth"'.format(self.interface,
                                                         self.vrid)
        rsp = self.device.cli_display(commands)
        by_line = rsp.split('\n')
        for each in by_line[1:]:
            if 'authentication-mode' in each:
                auth = each.split('authentication-mode')[-1].strip()
                auth_vars = auth.split(' ')
                if len(auth_vars) == 3:
                    auth_mode = auth_vars[0]
                    key_type = auth_vars[1]
                    key = auth_vars[2]
                    auth = dict(auth_mode=auth_mode, key_type=key_type,
                                key=key)
        return auth

    def _apply_value_maps(self, existing):
        """Manipulating value for preempt
        """

        if existing.get('preempt'):
            existing['preempt'] = existing.get('preempt').lower()

        return existing

    def remove(self, stage=False):
        """Stage or execute commands to remove VRRP group.

        Args:
            stage (bool): whether to stage the command or execute immediately

        Returns:
            True if stage=True and successfully staged
            CLI response string if immediate execution
        """
        commands = []
        commands.append(self.intf_command)
        commands.append('undo vrrp vrid {0}'.format(self.vrid))
        commands.append('\n')

        if stage:
            return self.device.stage_config(commands, 'cli_config')
        else:
            return self.device.cli_config(commands)

    def shutdown(self, stage=False):
        """Stage or execute commands to shutdown VRRP group.

        Args:
            stage (bool): whether to stage the command or execute immediately

        Returns:
            True if stage=True and successfully staged
            CLI response string if immediate execution
        """
        commands = []
        commands.append(self.intf_command)
        commands.append('vrrp vrid {0} shutdown'.format(self.vrid))
        commands.append('\n')

        if stage:
            return self.device.stage_config(commands, 'cli_config')
        else:
            return self.device.cli_config(commands)

    def undoshutdown(self, stage=False):
        """Stage or execute commands to undo shutdown of VRRP group.

        Args:
            stage (bool): whether to stage the command or execute immediately

        Returns:
            True if stage=True and successfully staged
            CLI response string if immediate execution
        """
        commands = []
        commands.append(self.intf_command)
        commands.append('undo vrrp vrid {0} shutdown'.format(self.vrid))
        commands.append('\n')

        if stage:
            return self.device.stage_config(commands, 'cli_config')
        else:
            return self.device.cli_config(commands)

    def get_config(self):
        """Get the config of a given vrid on a given interface

        Returns:
            This returns a dictionary with the following k/v pairs:
               :auth_mode (str): "simple" or "md5"
               :key_type (str): "cipher"
               :key (str): it will be a cipher
               :priority (str): VRRP priority
               :preempt (str): "true" or "false" (STRING)
               :vip (str): virtual IP address

        """

        self.get_vrrp_groups()
        config = [each for each in self._existing_all
                  if each.get('vrid') == self.vrid]
        if len(config) == 0:
            existing = {}
        elif len(config) == 1:
            existing = config[0]
            auth = self.get_auth_type()
            existing.update(auth)
            existing = self._apply_value_maps(existing)

        return existing

    def build(self, stage=False, **kvargs):
        """Execute or stage XML VRRP configuration and send to staging

        Args:
            stage (bool): whether to stage the command or execute immediately
            kvargs: see Keyword Args

        Note:
            If auth is being configured, all three auth Keyword
            Args are required.

        Keyword Args:
            vip (str): REQUIRED - virtual IP address
            auth_mode (str): OPTIONAL - "simple" or "md5"
            key_type (str): OPTIONAL - "cipher" or "plain"
            key (str): OPTIONAL - text string if ``key_type`` is "plain" or
                cipher if ``key_type`` is "cipher"
            priority (str): OPTIONAL - VRRP priority
            preempt (str): OPTIONAL - "yes" or "no" (STRING)

        Returns:
            True if stage=True and successfully staged
            CLI response if immediate execution
        """
        commands = []

        CMDS = {
            'priority': 'vrrp vrid {0} priority {1}',
            'preempt': 'vrrp vrid {0} preempt-mode',
            'vip': 'vrrp vrid {0} virtual-ip {1}',
            'auth': 'vrrp vrid {0} authentication-mode {1} {2} {3}',
        }

        vip = kvargs.get('vip')
        prio = kvargs.get('priority')
        preempt = kvargs.get('preempt')
        auth_mode = kvargs.get('auth_mode')
        key_type = kvargs.get('key_type')
        key = kvargs.get('key')

        if vip:
            commands.append((CMDS.get('vip')).format(self.vrid, vip))
        if prio:
            commands.append((CMDS.get('priority')).format(self.vrid, prio))
        if preempt == 'yes':
            commands.append(CMDS.get('preempt').format(self.vrid))
        elif preempt == 'no':
            commands.append('undo ' + CMDS.get('preempt').format(self.vrid))
        if auth_mode:
            commands.append((CMDS.get('auth')).format(self.vrid, auth_mode,
                                                      key_type, key))

        if commands:
            commands.insert(0, self.intf_command)
            commands.append('\n')
            if stage:
                return self.device.stage_config(commands, 'cli_config')
            else:
                return self.device.cli_config(commands)



"""Factory default HPCOM7 devices.
"""


class CleanErase(object):
    """Factory default a HP Comware 7 switch.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.

    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
    """

    def __init__(self, device):
        self.device = device

    def build(self, stage=False, factory_default=False):
        """Build cmd list to factory default switch and immediately reboot.

        Args:
            factory_default (bool): determines if the switch will be
                reset to factory defaults.  It is a safety measure and
                must be set to for the factory default to take place.
            stage (bool): whether to stage the command for later execution,
                or execute immediately.
        Returns:
            True if stage=True and staging is successful.
            The output of restore command if stage=False
        """
        if factory_default:
            commands = ['restore factory-default']
            if stage:
                return self.device.stage_config(commands, 'cli_display')
            else:
                try:
                    return self.device.cli_display(commands)
                finally:
                    self.device.reboot()



import unittest
import mock

from pyhpecw7.features.interface import Interface
from pyhpecw7.features.errors import InterfaceParamsError, InterfaceAbsentError, InterfaceTypeError

from .base_feature_test import BaseFeatureCase

IFACE_INDEX = '11'
LOGICAL_IFACE = 'LoopBack30'
ETH_IFACE = 'FortyGigE1/0/3'


class InterfaceTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    @mock.patch.object(Interface, '_get_iface_index')
    @mock.patch.object(Interface, '_is_ethernet_is_routed')
    def setUp(self, mock_is_eth, mock_get_index, mock_device):
        self.device = mock_device

        mock_get_index.return_value = IFACE_INDEX

        mock_is_eth.return_value = True, False
        self.eth_iface = Interface(self.device, ETH_IFACE)

        mock_is_eth.return_value = False, True
        self.lo_iface = Interface(self.device, LOGICAL_IFACE)

    def test_init(self):
        self.assertTrue(self.eth_iface.is_ethernet)
        self.assertFalse(self.eth_iface.is_routed)

        self.assertFalse(self.lo_iface.is_ethernet)
        self.assertTrue(self.lo_iface.is_routed)

        self.assertEqual(self.eth_iface.iface_index, IFACE_INDEX)
        self.assertEqual(self.lo_iface.iface_index, IFACE_INDEX)

        self.assertEqual(self.eth_iface.iface_type, 'FortyGigE')
        self.assertEqual(self.lo_iface.iface_type, 'LoopBack')

    def test_get_index(self):
        expected_get, get_reply = self.xml_get_and_reply('interface_index')
        self.device.get.return_value = get_reply

        expected = '9'
        result = self.eth_iface._get_iface_index()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    def test_is_eth_is_routed(self):
        expected_get, get_reply = self.xml_get_and_reply('interface_is_eth')
        self.device.get.return_value = get_reply

        expected = True, True
        result = self.eth_iface._is_ethernet_is_routed()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    def test_get_defaults(self):
        eth_defaults = self.eth_iface.get_default_config()
        lo_defaults = self.lo_iface.get_default_config()

        eth_expected = dict(description=ETH_IFACE + ' Interface', admin='up', speed='auto', duplex='auto', type='bridged')
        lo_expected = dict(description=LOGICAL_IFACE + ' Interface', admin='up', type='routed')

        self.assertEqual(eth_defaults, eth_expected)
        self.assertEqual(lo_defaults, lo_expected)

    def test_param_check(self):
        with self.assertRaises(InterfaceParamsError):
            self.lo_iface.param_check(speed='1000')

        self.eth_iface.iface_exists = False
        with self.assertRaises(InterfaceAbsentError):
            self.eth_iface.param_check(admin='up')

        self.eth_iface.iface_type = None
        with self.assertRaises(InterfaceTypeError):
            self.eth_iface.param_check(admin='up')

    def test_get_config(self):
        expected_get, get_reply = self.xml_get_and_reply('interface')
        self.device.get.return_value = get_reply

        expected = {'admin': 'up', 'duplex': 'auto', 'speed': 'auto', 'description': 'FortyGigE1/0/3 Interface', 'type': 'routed'}

        iface = self.eth_iface.get_config()

        self.assertEqual(iface, expected)
        self.assert_get_request(expected_get)

    def test_logical_iface(self):
        expected = self.read_action_xml('interface_logical')

        self.lo_iface._logical_iface()
        self.assert_action_request(expected)

        self.lo_iface._logical_iface(stage=True)
        self.assert_stage_request(expected, 'action')

    def test_logical_iface_remove(self):
        expected = self.read_action_xml('interface_logical_remove')
        self.lo_iface._logical_iface(remove=True)
        self.assert_action_request(expected)

    def test_build_config_default(self):
        expected = self.read_action_xml('interface_default')

        self.eth_iface._build_config('default', admin='up', duplex='half')
        self.assert_action_request(expected)

        self.eth_iface._build_config('default', stage=True, admin='up', duplex='half')
        self.assert_stage_request(expected, 'action')

    def test_build_config_present(self):
        expected = self.read_config_xml('interface')

        self.eth_iface._build_config('present', admin='up', duplex='half')
        self.assert_config_request(expected)

        self.eth_iface._build_config('present', stage=True, admin='up', duplex='half')
        self.assert_stage_request(expected, 'edit_config')

    def test_build_config_remove(self):
        expected = self.read_action_xml('interface_default')

        self.eth_iface._build_config('absent', admin='up', duplex='half')
        self.assert_action_request(expected)

        self.eth_iface._build_config('absent', stage=True, admin='up', duplex='half')
        self.assert_stage_request(expected, 'action')

    @mock.patch.object(Interface, '_logical_iface')
    def test_remove(self, mock_logical):
        self.lo_iface.remove_logical()
        mock_logical.assert_called_with(remove=True, stage=False)

        self.lo_iface.remove_logical(stage=True)
        mock_logical.assert_called_with(remove=True, stage=True)

    @mock.patch.object(Interface, '_logical_iface')
    def test_create(self, mock_logical):
        self.lo_iface.create_logical()
        mock_logical.assert_called_with(stage=False)

        self.lo_iface.create_logical(stage=True)
        mock_logical.assert_called_with(stage=True)

    @mock.patch.object(Interface, '_build_config')
    def test_build(self, mock_build):
        self.eth_iface.build(admin='up')
        mock_build.assert_called_with(state='present', stage=False, admin='up')

        self.eth_iface.build(admin='up', stage=True)
        mock_build.assert_called_with(state='present', stage=True, admin='up')

    @mock.patch.object(Interface, '_build_config')
    def test_default(self, mock_build):
        self.eth_iface.default()
        mock_build.assert_called_with(state='default', stage=False)

        self.eth_iface.default(stage=True)
        mock_build.assert_called_with(state='default', stage=True)

if __name__ == '__main__':
    unittest.main()



"""Manage switchports on HPCOM7 devices.
"""
from pyhpecw7.features.interface import Interface

from pyhpecw7.utils.xml.lib import *


class Switchport(object):
    """This class is used to get and build layer 2 interface
    configurations on ``HPCOM7`` devices.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        interface_name (str): The name of the interface.

    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        interface_name (str): The name of the interface.
        interface (pyhpecw7.features.Interface): The associated
            ``Interface`` configuration object.
    """
    def __init__(self, device, interface_name):
        self.device = device
        self.interface = Interface(device, interface_name)
        self.interface_name = self.interface.interface_name
        self.link_type = 'unknown'

    def get_default(self):
        """Return the default layer 2 settings for a switchport.

        Returns:
            A dictionary of default configuration parameters.

            For example::

                {
                    'pvid': '1',
                    'link_type': 'access'
                }
        """
        return {'pvid': '1', 'link_type': 'access'}

    def get_config(self):
        """Return the current layer 2 settings on the switchport.

        Returns:
            A dictionary of current configuration parameters.

            For example::

                {
                    'pvid': '2',
                    'link_type': 'trunk',
                    'permitted_vlans': '1-5'
                }
        """
        key_map = {'link_type': 'LinkType',
                   'permitted_vlans': 'PermitVlanList',
                   'pvid': 'PVID'}
        value_map = {'LinkType': {'1': 'access',
                                  '2': 'trunk'}}

        E = data_element_maker()
        top = E.top(
            E.VLAN(
                E.Interfaces(
                    E.Interface(
                        E.IfIndex(self.interface.iface_index)
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        reply_data = find_in_data('Interface', nc_get_reply.data_ele)

        if reply_data is None:
            return {}

        link_type_value = find_in_data('LinkType', nc_get_reply.data_ele).text

        if link_type_value == '1':
            self.link_type = 'access'
        elif link_type_value == '2':
            self.link_type = 'trunk'

        return data_elem_to_dict(reply_data, key_map, value_map=value_map)

    def convert_interface(self, link_type, stage=False):
        """Stage or execute the commands to toggle an interface between trunk/access.

        Args:
            link_type (str): 'access' or 'trunk'.
            stage (bool): whether to stage the command or execute immediately

        Note:
            If `link_type` does not equal 'access' or 'trunk',
            no commands are staged.

        Returns:
            True if stage=True and successfully staged
            etree.Element XML responses if immediate execution
        """
        if link_type == 'access':
            type_value = '1'
        elif link_type == 'trunk':
            type_value = '2'
        else:
            return

        EN = nc_element_maker()
        EC = config_element_maker()
        config = EN.config(
            EC.top(
                EC.Ifmgr(
                    EC.Interfaces(
                        EC.Interface(
                            EC.IfIndex(self.interface.iface_index),
                            EC.LinkType(type_value),
                        )
                    )
                )
            )
        )

        if stage:
            self.device.stage_config(config, 'edit_config')
        else:
            self.device.edit_config(config)

    def default(self, stage=False):
        """Stage or execute a layer 2 default configuration.

        Args:
            stage (bool): whether to stage the command or execute immediately

        Returns:
            True if stage=True and successfully staged
            etree.Element XML response if immediate execution
        """
        defaults = self.get_default()
        return self.build(stage=stage, **defaults)

    def build(self, stage=False, **params):
        """Stage a layer 2 configuration with given parameters on switchport.

        Args:
            stage (bool): whether to stage the command or execute immediately

        Keyword Args:
            link_type (str): 'access' or 'trunk'.
            pvid (str): The access VLAN if link_type is 'access',
                the native VLAN if link_type is 'trunk'.
            permitted_vlans (str): A comma and/or hyphen delimited list
                of VLAN numbers. Used when link_type is 'trunk'.
                For example: `1,3-5,7`

        Returns:
            True if stage=True and successfully staged
            etree.Element XML response if immediate execution
        """
        EN = nc_element_maker()
        EC = config_element_maker()

        link_type = params.pop('link_type', None)
        if link_type == 'trunk':
            if self.link_type != 'trunk':
                self.convert_interface(link_type, stage=stage)

            key_map = {'permitted_vlans': 'PermitVlanList',
                       'pvid': 'PVID'}

            config = EN.config(
                EC.top(
                    EC.VLAN(
                        EC.TrunkInterfaces(
                            EC.Interface(
                                EC.IfIndex(self.interface.iface_index),
                                *config_params(params, key_map)
                            )
                        )
                    )
                )
            )

        elif link_type == 'access':
            if self.link_type != 'access':
                self.convert_interface(link_type, stage=stage)

            key_map = {'pvid': 'PVID'}
            config = EN.config(
                EC.top(
                    EC.VLAN(
                        EC.AccessInterfaces(
                            EC.Interface(
                                EC.IfIndex(self.interface.iface_index),
                                *config_params(params, key_map)
                            )
                        )
                    )
                )
            )

        else:
            return

        if params:
            if stage:
                return self.device.stage_config(config, 'edit_config')
            else:
                return self.device.edit_config(config)

        return False



import unittest
import mock

from pyhpecw7.features.portchannel import Portchannel, InvalidPortType, AggregationGroupError
from .base_feature_test import BaseFeatureCase

R_GROUP_ID = '101'
B_GROUP_ID = '100'

class PortChannelTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self,  mock_device):
        self.device = mock_device
        self.rpc = Portchannel(self.device, R_GROUP_ID, 'routed')
        self.bpc = Portchannel(self.device, B_GROUP_ID, 'bridged')

    def test_init(self):
        self.assertEqual(self.rpc.pc_type, 'routed')
        self.assertEqual(self.bpc.pc_type, 'bridged')


    def test_get_portchannels(self):
        expected_get, get_reply = self.xml_get_and_reply('portchannel_list')
        self.device.get.return_value = get_reply

        expected = ['3', '100']

        result = self.bpc.get_portchannels()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch.object(Portchannel, 'get_interface_from_index')
    @mock.patch.object(Portchannel, 'get_lacp_mode_by_name')
    def test_get_config(self, mock_lacp_mode, mock_iface_from_index):
        mock_lacp_mode.return_value = 'active'

        iface_list = ['FortyGigE1/0/2', 'FortyGigE1/0/1']
        def mock_iface_gen(self):
            return iface_list.pop()

        mock_iface_from_index.side_effect = mock_iface_gen

        expected_get, get_reply = self.xml_get_and_reply('portchannel')
        self.device.get.return_value = get_reply

        expected = {
            'min_ports': None,
            'max_ports': None,
            'lacp_modes_by_interface': [
                {
                    'interface': 'FortyGigE1/0/1',
                    'lacp_mode': 'active'
                },
                {
                    'interface': 'FortyGigE1/0/2',
                    'lacp_mode': 'active'
                }
            ],
            'nc_groupid': '100',
            'mode': 'static',
            'members': [
                'FortyGigE1/0/1',
                'FortyGigE1/0/2'
            ],
            'lacp_edge': 'disabled',
            'pc_index': '33736',
            'groupid': '100'
        }

        result = self.bpc.get_config()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch.object(Portchannel, 'get_interface_from_index')
    def test_all_members(self, mock_iface_from_index):
        iface_list = ['FortyGigE1/0/2', 'FortyGigE1/0/1']
        def mock_iface_gen(self):
            return iface_list.pop()

        mock_iface_from_index.side_effect = mock_iface_gen

        expected_get, get_reply = self.xml_get_and_reply('portchannel_all_members')
        self.device.get.return_value = get_reply

        expected = ['FortyGigE1/0/1', 'FortyGigE1/0/2']
        result = self.bpc.get_all_members()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch.object(Portchannel, 'get_interface_from_index')
    def test_all_members_ifindex(self, mock_iface_from_index):
        iface_list = ['FortyGigE1/0/2', 'FortyGigE1/0/1']
        def mock_iface_gen(self):
            return iface_list.pop()

        mock_iface_from_index.side_effect = mock_iface_gen

        expected_get, get_reply = self.xml_get_and_reply('portchannel_all_members')
        self.device.get.return_value = get_reply

        expected = ['1', '5']
        result = self.bpc.get_all_members(list_type='ifindex')

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch.object(Portchannel, 'get_interface_from_index')
    def test_all_members_asdict(self, mock_iface_from_index):
        iface_list = ['FortyGigE1/0/2', 'FortyGigE1/0/1']
        def mock_iface_gen(self):
            return iface_list.pop()

        mock_iface_from_index.side_effect = mock_iface_gen

        expected_get, get_reply = self.xml_get_and_reply('portchannel_all_members')
        self.device.get.return_value = get_reply

        expected = {'FortyGigE1/0/2': '100', 'FortyGigE1/0/1': '100'}
        result = self.bpc.get_all_members(asdict=True)

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch.object(Portchannel, 'get_index_from_interface')
    def test_lacp_mode_by_name_passive(self, mock_index_from_iface):
        mock_index_from_iface.return_value = '41'

        expected_get, get_reply = self.xml_get_and_reply('portchannel_lacp_mode_by_name_passive')
        self.device.get.return_value = get_reply

        expected = 'passive'
        result = self.bpc.get_lacp_mode_by_name('FortyGigE1/0/11')

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch.object(Portchannel, 'get_index_from_interface')
    def test_lacp_mode_by_name_active(self, mock_index_from_iface):
        mock_index_from_iface.return_value = '1'

        expected_get, get_reply = self.xml_get_and_reply('portchannel_lacp_mode_by_name_active')
        self.device.get.return_value = get_reply

        expected = 'active'
        result = self.bpc.get_lacp_mode_by_name('FortyGigE1/0/1')

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    def test_interface_from_index(self):
        expected_get, get_reply = self.xml_get_and_reply('portchannel_interface_from_index')
        self.device.get.return_value = get_reply

        expected = 'FortyGigE1/0/11'
        result = self.bpc.get_interface_from_index('41')

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch('pyhpecw7.features.portchannel.Interface')
    def test_index_from_Interface(self, mock_iface):
        mock_iface.return_value.iface_index = '96'

        result = self.bpc.get_index_from_interface('FortyGigE1/0/1')
        self.assertEqual(result, '96')

        # check local caching
        mock_iface.return_value.iface_index = '1111'

        result = self.bpc.get_index_from_interface('FortyGigE1/0/1')
        self.assertEqual(result, '96')

    @mock.patch.object(Portchannel, 'get_index_from_interface')
    def test_add_lagg_member(self, mock_index_from_iface):
        mock_index_from_iface.return_value = '1'
        expected = self.read_config_xml('portchannel_lagg_member')
        result = self.bpc._add_lagg_member('FortyGigE1/0/1', lacp='active')

        self.assert_elements_equal(result, expected)

    @mock.patch.object(Portchannel, 'get_index_from_interface')
    def test_build_config(self, mock_index_from_iface):
        index_list = ['1', '5']
        def mock_index_gen(interface):
            return index_list.pop(0)

        mock_index_from_iface.side_effect = mock_index_gen

        expected_call = self.read_config_xml('portchannel_basic')
        self.bpc._build_config('present', members=['FortyGigE1/0/1', 'FortyGigE1/0/2'])

        self.assert_config_request(expected_call)

    @mock.patch.object(Portchannel, 'get_index_from_interface')
    def test_build_config_stage(self, mock_index_from_iface):
        index_list = ['1', '5']
        def mock_index_gen(interface):
            return index_list.pop(0)

        mock_index_from_iface.side_effect = mock_index_gen

        expected_call = self.read_config_xml('portchannel_basic')
        self.bpc._build_config('present', stage=True, members=['FortyGigE1/0/1', 'FortyGigE1/0/2'])

        self.assert_stage_request(expected_call, 'edit_config')

    @mock.patch.object(Portchannel, 'get_index_from_interface')
    def test_build_config_remove_members(self, mock_index_from_iface):
        index_list = ['1', '5', '9']
        def mock_index_gen(interface):
            return index_list.pop(0)

        mock_index_from_iface.side_effect = mock_index_gen

        self.bpc.members_to_remove = ['FortyGigE1/0/3']

        expected_call = self.read_config_xml('portchannel_basic_remove_members')
        self.bpc._build_config('present', members=['FortyGigE1/0/1', 'FortyGigE1/0/2'])

        self.assert_config_request(expected_call)

    @mock.patch.object(Portchannel, 'get_index_from_interface')
    def test_build_config_change_lacp(self, mock_index_from_iface):
        index_list = ['1', '5', '9']
        def mock_index_gen(interface):
            return index_list.pop(0)

        mock_index_from_iface.side_effect = mock_index_gen

        self.bpc.desired_lacp_mode = 'bridged'

        expected_call = self.read_config_xml('portchannel_change_lacp')
        self.bpc._build_config('present', members=['FortyGigE1/0/1', 'FortyGigE1/0/2'], lacp_to_change=['FortyGigE1/0/1'])

        self.assert_config_request(expected_call)

    def test_build_config_no_members(self):
        expected_call = self.read_config_xml('portchannel_no_members')
        self.bpc._build_config('present')

        self.assert_config_request(expected_call)

    def test_build_config_remove(self):
        expected_call = self.read_config_xml('portchannel_remove')
        self.bpc._build_config('absent')

        self.assert_config_request(expected_call)

    def test_build_config_min_max(self):
        expected_xml_call = self.read_config_xml('portchannel_no_members')
        expected_cli_call = ['interface Bridge-Aggregation 100', 'link-aggregation selected-port minimum 2', 'link-aggregation selected-port maximum 8']

        self.bpc._build_config('present', min_ports='2', max_ports='8')

        self.assert_config_request(expected_xml_call)
        self.bpc.device.cli_config.assert_called_with(expected_cli_call)

    @mock.patch.object(Portchannel, '_build_config')
    def test_build(self, mock_build_config):
        kwargs = dict(members=['FortyGigE1/0/1', 'FortyGigE1/0/2'])
        self.bpc.build(**kwargs)
        mock_build_config.assert_called_with(state='present', stage=False, **kwargs)

    @mock.patch.object(Portchannel, '_build_config')
    def test_remove(self, mock_build_config):
        self.bpc.remove()
        mock_build_config.assert_called_with(state='absent', stage=False)

    @mock.patch('pyhpecw7.features.portchannel.Interface')
    @mock.patch.object(Portchannel, 'get_all_members')
    def test_param_check(self, mock_get_all_members, mock_iface):
        mock_iface.return_value.get_config.return_value = {'type': 'bridged'}
        mock_get_all_members.return_value = {'FortyGigE1/0/1': B_GROUP_ID}
        self.bpc.param_check(members=['FortyGigE1/0/1'])

        mock_iface.return_value.get_config.return_value = {'type': 'routed'}
        with self.assertRaises(InvalidPortType):
            self.bpc.param_check(members=['FortyGigE1/0/1'])

        mock_iface.return_value.get_config.return_value = {'type': 'bridged'}
        mock_get_all_members.return_value = {'FortyGigE1/0/1': '300'}
        with self.assertRaises(AggregationGroupError):
            self.bpc.param_check(members=['FortyGigE1/0/1'])


if __name__ == '__main__':
    unittest.main()



"""Manage interfaces on HPCOM7 devices.
"""

from pyhpecw7.utils.xml.lib import reverse_value_map
from pyhpecw7.features.errors import InterfaceCreateError, InterfaceTypeError,\
    InterfaceAbsentError, InterfaceParamsError, InterfaceVlanMustExist
from pyhpecw7.features.vlan import Vlan

from pyhpecw7.utils.xml.lib import *


class Interface(object):
    """This class is used to get
    and build interface configurations on ``HPCOM7`` devices.

    Args:
        device (HPCOM7): connected instance of a
            ``phyp.comware.HPCOM7`` object.
        interface_name (str): The name of the interface.

    Attributes:
        device (HPCOM7): connected instance of a
            ``phyp.comware.HPCOM7`` object.
        interface_name (str): The name of the interface.
        iface_index (str): The device's internal number representation
            of an interface.
        iface_type (str): The type of interface,
            for example: 'LoopBack', 'FortyGigE'.
        is_ethernet (bool): Whether the interface is ethernet.
        is_routed (bool): Whether the interface is in layer 3 mode.
            If this is ``False``, the interface is either in bridged
            mode or does not exist.
        iface_exists (bool): Whether the interface exists. Physical
            interfaces should always exist. Logical interfaces may
            or may not exist.
    """
    def __init__(self, device, interface_name):
        # used to map key values from our dictionary model
        # to expected XML tags and vice versa
        self._key_map = {
            'admin': 'AdminStatus',
            'speed': 'ConfigSpeed',
            'duplex': 'ConfigDuplex',
            'description': 'Description',
            'type': 'PortLayer'
        }

        # used to map value values from our dictionary model
        # to expected XML tags and vice versa
        self._value_map = {
            'AdminStatus': {'1': 'up',
                            '2': 'down'},
            'ConfigSpeed': {'1': 'auto', '2': '10',
                            '4': '100', '32': '1000',
                            '1024': '10000', '4096': '20000',
                            '8192': '40000', '16384': '100000'},
            'ConfigDuplex': {'1': 'full',
                             '2': 'half',
                             '3': 'auto'},
            'PortLayer': {'1': 'bridged',
                          '2': 'routed'}
        }

        self._iface_types = set(['FortyGigE', 'Tunnel', 'LoopBack',
                                 'Vlan-interface', 'Bridge-Aggregation',
                                 'Route-Aggregation', 'GigabitEthernet',
                                 'TwentyGigE', 'Ten-GigabitEthernet',
                                 'HundredGigE'])

        # xml tags
        self._iface_row_name = 'Interface'
        self._iface_index_name = 'IfIndex'

        # usd in conjunction with key map and value map above
        self._r_key_map = dict(reversed(item) for item in self._key_map.items())
        self._r_value_map = reverse_value_map(self._r_key_map, self._value_map)

        # connect to the device and get more information
        self.interface_name, self.iface_type = self._iface_type(interface_name)
        self.device = device

        # The interface index is needed for most interface NETCONF requests
        self.iface_index = self._get_iface_index()

        self.is_ethernet, self.is_routed = self._is_ethernet_is_routed()
        self.iface_exists = True if self.iface_index else False

    def _iface_type(self, if_name):
        """Return the normalized interface name and type
        from a denormalized interface name.
        """

        if if_name.lower().startswith('gi'):
            if_type = 'GigabitEthernet'
        elif if_name.lower().startswith('ten'):
            if_type = 'Ten-GigabitEthernet'
        elif if_name.lower().startswith('fo'):
            if_type = 'FortyGigE'
        elif if_name.lower().startswith('vl'):
            if_type = 'Vlan-interface'
        elif if_name.lower().startswith('lo'):
            if_type = 'LoopBack'
        elif if_name.lower().startswith('br'):
            if_type = 'Bridge-Aggregation'
        elif if_name.lower().startswith('ro'):
            if_type = 'Route-Aggregation'
        elif if_name.lower().startswith('tu'):
            if_type = 'Tunnel'
        elif if_name.lower().startswith('tw'):
            if_type = 'TwentyGigE'
        elif if_name.lower().startswith('hu'):
            if_type = 'HundredGigE'
        else:
            if_type = None

        number_list = if_name.split(' ')
        if len(number_list) == 2:
            number = number_list[-1].strip()
        else:
            number = self._get_number(if_name)

        if if_type:
            proper_interface = if_type + number
        else:
            proper_interface = if_name

        return proper_interface, if_type

    def _get_number(self, if_name):
        digits = ''
        for char in if_name:
            if char.isdigit() or char == '/' or char == ':':
                digits += char
        return digits

    def _get_iface_index(self):
        """Return the interface index given the self.interface_name
        attribute by asking the device. If the interface doesn't exist,
        return the empty string.
        """
        E = data_element_maker()
        top = E.top(
            E.Ifmgr(
                E.Interfaces(
                    E.Interface(
                        E.Name(self.interface_name)
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        reply_data = find_in_data(
            self._iface_index_name, nc_get_reply.data_ele)

        if reply_data is None:
            return ''

        return reply_data.text

    def _is_ethernet_is_routed(self):
        """Return whether the interface is ethernet and whether
        it is routed. If the interface doesn't exist,
        return False.
        """
        E = data_element_maker()
        top = E.top(
            E.Ifmgr(
                E.Interfaces(
                    E.Interface(
                        E.IfIndex(self.iface_index)
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        reply_data = find_in_data('ifType', nc_get_reply.data_ele)

        routed_reply_data = find_in_data('PortLayer', nc_get_reply.data_ele)

        is_ethernet = False
        is_routed = False
        try:
            if reply_data.text == '6':
                is_ethernet = True
        except AttributeError:
            pass

        try:
            if routed_reply_data.text == '2':
                is_routed = True
        except AttributeError:
            pass

        return is_ethernet, is_routed

    def update(self):
        """Update ``self.iface_index`` and ``self.iface_exists``.

        Usually called after a logical interface is created.

        Raises:
            InterfaceCreateError: if the interface hasn't yet
                been successfully created.

        Note:
            It is the responsibility of the caller to call ``update()`
            after staging (``create_logical()``) *and* executing
            (``execute()`` on this class's ``device`` object) of
            commands to create an interface.
         """
        if_index = self._get_iface_index()
        if not if_index:
            raise InterfaceCreateError(self.interface_name)
        self.iface_index = if_index
        self.iface_exists = True

    def get_default_config(self):
        """Return the default configuration of an interface.

        Returns:
            A dictionary of default interface configuration parameters,
            depending on the type of interface.

            For example, for ethernet interfaces::

                {
                    'description': 'FortyGigE1/0/1 Interface',
                    'admin': 'up',
                    'speed': 'auto',
                    'duplex': 'auto',
                    'type': 'bridged'
                }
        """
        if not self.iface_type:
            return None

        defaults = {}
        defaults['description'] = self.interface_name + ' Interface'
        defaults['admin'] = 'up'
        if self.is_ethernet:
            defaults['speed'] = 'auto'
            defaults['duplex'] = 'auto'
            defaults['type'] = 'bridged'
        elif self.iface_type == 'Bridge-Aggregation':
            defaults['type'] = 'bridged'
        else:
            defaults['type'] = 'routed'

        return defaults

    def param_check(self, **params):
        """Checks given parameters against the interface for various errors.

        Args:
            **params: see Keyword Args

        Keyword Args:
            admin (str): The up/down state of the interface.
                'up' or 'down'.
            speed (str): The speed of the interface, in Mbps.
            duplex (str): The duplex of the interface.
                'full', 'half', or 'auto'.
            description (str): The textual description of the interface.
            type (str): Whether the interface is in layer 2 or layer 3 mode.
                'bridged' or 'routed'.

        Raises:
            InterfaceTypeError: if the given interface isn't a valid type.
            InterfaceAbsentError: if the given interface is of type is_ethernet
                and doesn't exist.
            InterfaceParamsError: if 'speed' or 'duplex' are supplied for a
                non ethernet interface.
            InterfaceVlanMustExist: if the interface is of type
                'Vlan-interface' and the the associated vlan doesn't exist.
        """
        if not self.iface_type:
            raise InterfaceTypeError(
                self.interface_name, list(self._iface_types))

        if not self.iface_exists:
            if self.iface_type in {'FortyGigE', 'GigabitEthernet',
                                   'TwentyGigE', 'Ten-GigabitEthernet',
                                   'HundredGigE'}:
                raise InterfaceAbsentError(self.interface_name)

        if not self.is_ethernet:
            param_names = []
            if params.get('speed'):
                param_names.append('speed')
            if params.get('duplex'):
                param_names.append('duplex')
            if param_names:
                raise InterfaceParamsError(self.interface_name, param_names)

        if self.iface_type == 'Vlan-interface':
            number = self.interface_name.split('Vlan-interface')[1]
            vlan = Vlan(self.device, number)
            if not vlan.get_config():
                raise InterfaceVlanMustExist(self.interface_name, number)

    def get_config(self):
        """Return the currently configured
        parameters for the interface.

        Returns:
            A dictionary of currently configured
            parameters for the interface, including:
                :admin (str): The up/down state of the interface.
                    'up' or 'down'.
                :speed (str): The speed of the interface, in Mbps.
                :duplex (str): The duplex of the interface.
                    'full', 'half', or 'auto'.
                :description (str): The textual description of the interface.
                :type (str): Whether the interface is in layer 2 or
                    layer 3 mode. 'bridged' or 'routed'.
        """
        E = data_element_maker()
        top = E.top(
            E.Ifmgr(
                E.Interfaces(
                    E.Interface(
                        E.IfIndex(self.iface_index)
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        reply_data = find_in_data(self._iface_row_name, nc_get_reply.data_ele)

        if reply_data is None:
            return {}

        return data_elem_to_dict(reply_data, self._key_map, value_map=self._value_map)

    def create_logical(self, stage=False):
        """Stage or execute the configuration to create
        a logical interface.

        Supported types include 'LoopBack',
        'Vlan-interface', 'Bridge-Aggregation',
        and 'Route-Aggregation'

        Note:
            When stage=True, it's the caller's responsibility to call
            ``execute()`` on this class's ``device``
            object after this method is called.

        Note:
            After execution, the caller must call ``update()`` on this class.

        Returns:
            True if successful.

        Raises:
            InterfaceCreateError: if the logical interface
                cannot be created.
        """
        return self._logical_iface(stage=stage)

    def remove_logical(self, stage=False):
        """Stage or execute the configuration to remove
        a logical interface.

        Supported types include 'LoopBack',
        'Vlan-interface', 'Bridge-Aggregation',
        and 'Route-Aggregation'

        Args:
            stage (bool): whether to stage the commands or execute
                immediately

        Note:
            It's the caller's responsibility to call
            ``execute()`` on this class's ``device``
            object after this method is called.

        Returns:
            True if stage=True and staging is successful
            etree.Element XML response if immediate execution

        Raises:
            InterfaceCreateError: if the logical interface
                cannot be removed.
        """
        return self._logical_iface(remove=True, stage=stage)

    def _logical_iface(self, remove=False, stage=False):
        """Stage or execute the configuration to create
        or remove a logical interface.

        Args:
            remove (bool): If ``True``, the logical
                interface is removed. If ``False``,
                the logical interface is created.

            stage (bool): whether to stage the commands or execute
                immediately

        Returns:
            True if stage=True and staging is successful
            etree.Element XML response if immediate execution
        """
        logic_type_map = {'LoopBack': '16',
                          'Vlan-interface': '41',
                          'Bridge-Aggregation': '56',
                          'Route-Aggregation': '67'}

        if self.iface_type not in logic_type_map:
            raise InterfaceCreateError(self.interface_name)

        iface_number = self.interface_name.split(self.iface_type)[1]

        E = action_element_maker()
        top = E.top(
            E.Ifmgr(
                E.LogicInterfaces(
                    E.Interface(
                        E.IfTypeExt(logic_type_map[self.iface_type]),
                        E.Number(iface_number)
                    )
                )
            )
        )

        if remove:
            find_in_action('Interface', top).append(E.Remove())

        if stage:
            return self.device.stage_config(top, 'action')
        else:
            return self.device.action(top)

    def build(self, stage=False, **params):
        """Stage or execute the configuration to
        modify an interface.

        Args:
            stage (bool): whether to stage the commands or execute
                immediately
            **params: see Keyword Args.

        Keyword Args:
            admin (str): The up/down state of the interface.
                'up' or 'down'.
            speed (str): The speed of the interface, in Mbps.
            duplex (str): The duplex of the interface.
                'full', 'half', or 'auto'.
            description (str): The textual description of the interface.
            type (str): Whether the interface is in layer 2 or layer 3 mode.
                'bridged' or 'routed'.

        Raises:
            InterfaceCreateError: if a logical interface cannot be created.

        Returns:
            True if stage=True and staging is successful
            etree.Element XML response if immediate execution
        """
        return self._build_config(state='present', stage=stage, **params)

    def default(self, stage=False):
        """Stage or execute the configuration to default an interface.

        stage (bool): whether to stage the commands or execute
            immediately

        Returns:
            True if stage=True and staging is successful
            etree.Element XML response if immediate execution
        """
        return self._build_config(state='default', stage=stage)

    def _build_config(self, state, stage=False, **params):
        """Stage or execute the configuration to
        configure, default, or remove an interface.

        Args:
            state (str): 'present' configures,
                'absent' defaults,
                'default' defaults.
            stage (bool): whether to stage the commands or execute
                immediately
            **params: Used when state=present, see Keyword Args.

        Keyword Args:
            admin (str): The up/down state of the interface.
                'up' or 'down'.
            speed (str): The speed of the interface, in Mbps.
            duplex (str): The duplex of the interface.
                'full', 'half', or 'auto'.
            description (str): The textual description of the interface.
            type (str): Whether the interface is in layer 2 or layer 3 mode.
                'bridged' or 'routed'.

        Returns:
            True if stage=True and staging is successful
            etree.Element XML response if immediate execution
            False if illegal operation, e.g. removing a physical interface
        """
        if state == 'default':
            if self.iface_exists:
                E = action_element_maker()
                top = E.top(
                    E.Ifmgr(
                        E.Interfaces(
                            E.Interface(
                                E.IfIndex(self.iface_index),
                                E.Default()
                            )
                        )
                    )
                )

                if stage:
                    return self.device.stage_config(top, 'action')
                else:
                    return self.device.action(top)

        if state == 'present':
            params[self._iface_index_name] = self.iface_index

            EN = nc_element_maker()
            EC = config_element_maker()

            config = EN.config(
                EC.top(
                    EC.Ifmgr(
                        EC.Interfaces(
                            EC.Interface(
                                *config_params(params, self._key_map, value_map=self._r_value_map)
                            )
                        )
                    )
                )
            )

            if stage:
                return self.device.stage_config(config, 'edit_config')
            else:
                return self.device.edit_config(config)

        if state == 'absent':
            if self.is_ethernet:
                return self._build_config('default', stage=stage)

        return False



"""Manage VXLAN configurations on HPCOM7 devices.
"""
from lxml import etree
from pyhpecw7.utils.xml.lib import *
from pyhpecw7.utils.templates import cli
from pyhpecw7.features.interface import Interface


class Tunnel(object):
    """This class is used to get data and configure VXLAN tunnel interfaces.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        tunnel (str): Tunnel ID

    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        tunnel (str): Tunnel ID

    """
    def __init__(self, device, tunnel):

        self.device = device
        self.tunnel = tunnel

        # mapping of dictionary keys to XML tags
        self.TUNNELS = {
            'tunnel': 'TunnelID',
            'vxlan': 'VxlanID'
        }

    def get_config(self):
        """Get running config for a tunnel interface

        Returns:
            A dictionary is returned with the following k/v pairs:
                src (str): source IP addr of tunnel
                dest (str): destination IP addr of tunnel
                mode (str): mode of tunnel

            If the tunnel does not exist, an empty dictionary is returned.

        """
        existing = {}
        config = self.device.cli_display(
            'display current-configuration interface Tunnel {0}'.format(
                self.tunnel))

        parsed = cli.get_structured_data('tunnel.tmpl', config)

        if not parsed:
            existing = {}  # i.e, does not exist
        else:
            if len(parsed) == 1:  # which it should!
                existing = parsed[0]

        return existing

    def get_global_source(self):
        """Get global source address for tunnel interfaces

        Returns:
            String that is the global source IP address on the switch
        """
        address = None
        config = self.device.cli_display(
            'display current-configuration | inc "tunnel global source"')
        config_list = config.split('\n')
        if len(config_list) == 2:
            return None
        else:
            for each in config_list:
                if 'tunnel global' in each:
                    address = each.split(
                        'tunnel global source-address')[-1].strip()
        return address

    def build(self, stage=False, **kvargs):
        """Stage or execute config object to create/update tunnel

        Args:
            stage (bool): whether to stage the command or execute immediately

        Returns:
            True if stage=True and successfully staged
            CLI response strings if immediate execution
        """
        return self._build_config(state='present', stage=stage, **kvargs)

    def remove(self, stage=False):
        """Build config object to remove tunnel interface

        Args:
            stage (bool): whether to stage the command or execute immediately

        Returns:
            True if stage=True and successfully staged
            CLI response strings if immediate execution
        """
        return self._build_config(state='absent', stage=stage)

    def _build_config(self, state, stage=False, **kvargs):
        """Build CLI commands to configure/create VXLAN tunnel interfaces

        Args:
            state (str): must be "absent" or "present"
            kvargs: see Keyword Args
            stage (bool): whether to stage the command or execute immediately

        Keyword Args:
            src (str): OPTIONAL - source IP addr of tunnel
            dest (str): OPTIONAL - destination IP addr of tunnel
            global_src (str): OPTIONAL - global src IP addr for tunnels

        Returns:
            True if stage=True and successfully staged
            CLI response strings if immediate execution
        """
        commands = []
        if state == 'absent':
            commands.append('undo interface tunnel {0}'.format(self.tunnel))
        elif state == 'present':
            if kvargs.get('global_src'):
                commands.append('tunnel global source-address {0}'.format(
                    kvargs.get('global_src')))
            # has a not kvargs because tunnel ID won't be in there
            # if it's a new tunnel when using Ansible
            if kvargs.get('src') or kvargs.get('dest') or not kvargs:
                commands.append('interface tunnel {0} mode vxlan'.format(
                    self.tunnel))
                if kvargs.get('src'):
                    commands.append('source {0}'.format(kvargs.get('src')))
                if kvargs.get('dest'):
                    commands.append('destination {0}'.format(
                        kvargs.get('dest')))

        if commands:
            if stage:
                self.device.stage_config(commands, 'cli_config')
            else:
                self.device.cli_config(commands)


class Vxlan(object):
    """This class is used to get data and configure VXLAN/VSI mappings.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        vxlan (str): VXLAN ID
        vsi (str): name of the VSI

    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        vxlan (str): VXLAN ID
        vsi (str): name of the VSI

    """
    def __init__(self, device, vxlan, vsi=None):

        self.device = device
        self.vxlan = vxlan
        self.vsi = vsi
        self.VSI = {
            'vsi': 'VsiName',
            'descr': 'Description',
        }

        # mapping of dictionary keys to XML tags
        self.VXLAN = {
            'vxlan': 'VxlanID',
            'vsi': 'VsiName',
        }
        # mapping of dictionary keys to XML tags
        self.TUNNELS = {
            'tunnel': 'TunnelID',
            'vxlan': 'VxlanID'
        }

    def get_config(self):
        """Get associated VSI for a given VXLAN ID along with configured
        tunnels for that given VXLAN/VSI mapping.

        Returns:
            Dictionary with the following key/value pairs:
                :vxlan (str): vxlan id
                :vsi (str): name of vsi

            If the mapping does not exist, an empty dictionary is returned.

        """
        E = data_element_maker()
        top = E.top(
            E.VXLAN(
                E.VXLANs(
                    E.Vxlan(
                        E.VxlanID(self.vxlan)
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        return_vxlan = data_elem_to_dict(nc_get_reply.data_ele, self.VXLAN)

        if return_vxlan:
            return_vxlan['tunnels'] = self.get_tunnels()

        return return_vxlan

    def _build_vsi(self, operation, vsi=None):
        """Builds object to create/delete a VSI

        Args:
            operation (str): "merge" or "delete"
            kvargs:

        Keyword Args:
            vsi (str): name of VSI

        Returns:
            etree.Element config object to create/delete VSI
        """
        if not vsi:
            vsi = self.vsi

        EN = nc_element_maker()
        EC = config_element_maker()

        config = EN.config(
            EC.top(
                EC.L2VPN(
                    EC.VSIs(
                        EC.VSI(
                            EC.VsiName(vsi)
                        )
                    ),
                    **operation_kwarg(operation)
                )
            )
        )

        return config

    def _build_vxlan(self, operation):
        """Creates object to create a VXLAN and associate to a VSI name

        Args:
            operation (str): "merge" or "delete"
        """
        EN = nc_element_maker()
        EC = config_element_maker()

        vsi_name_ele = []
        if operation == 'merge':
            vsi_name_ele.append(EC.VsiName(self.vsi))

        config = EN.config(
            EC.top(
                EC.VXLAN(
                    EC.VXLANs(
                        EC.Vxlan(
                            EC.VxlanID(self.vxlan),
                            *vsi_name_ele
                        )
                    ),
                    **operation_kwarg(operation)
                )
            )
        )

        return config

    def _build_tunnels(self, operation, tunnels):
        """Updates tunnel objects for a given VXLAN/VSI

            Args:
                operation (str): "merge" or "delete"
                tunnels (list): list of tunnel ID (strings) to be modified
        """
        EN = nc_element_maker()
        EC = config_element_maker()

        tunnel_eles = []
        for tunnel in tunnels:
            tunnel_eles.append(EC.Tunnel(EC.TunnelID(tunnel), EC.VxlanID(self.vxlan)))

        config = EN.config(
            EC.top(
                EC.VXLAN(
                    EC.Tunnels(*tunnel_eles),
                    **operation_kwarg(operation)
                )
            )
        )

        return config

    def build(self, stage=False, **kvargs):
        """Stage or execute config for managing tunnels

        Args:
            state (str): "present" or "absent"
            kvargs: see Keyword Args
            stage (bool): whether to stage the command or execute immediately

        Keyword Args:
            tunnels_to_add (list): OPTIONAL - tunnels to add to the VXLAN/VSI
                mapping
            tunnels_to_remove (list): OPTIONAL - tunnels to remove to the
                VXLAN/VSI mapping

        Returns:
            True if stage=True and successfully staged
            List of etree.Element XML responses if immediately executed
        """
        rmv = True
        add = True

        if kvargs.get('tunnels_to_remove'):
            rmv_config = self._build_tunnels('delete', kvargs.get('tunnels_to_remove'))
            if stage:
                rmv = self.device.stage_config(rmv_config, 'edit_config')
            else:
                rmv = self.device.edit_config(rmv_config)

        if kvargs.get('tunnels_to_add'):
            add_config = self._build_tunnels('merge', kvargs.get('tunnels_to_add'))
            if stage:
                add = self.device.stage_config(add_config, 'edit_config')
            else:
                add = self.device.edit_config(add_config)

        if stage:
            return rmv and add
        else:
            return [rmv, add]

    def create(self, stage=False):
        """Stage or execute a config for creating a VSI

        Args:
            stage (bool): whether to stage the command or execute immediately

        Returns:
            True if stage=True and successfully staged
            List of etree.Element XML responses if immediately executed
        """
        vsi_config = self._build_vsi('merge')
        vxlan_config = self._build_vxlan('merge')

        if stage:
            vsi = self.device.stage_config(vsi_config, 'edit_config')
            vxlan = self.device.stage_config(vxlan_config, 'edit_config')
        else:
            vsi = self.device.edit_config(vsi_config)
            vxlan = self.device.edit_config(vxlan_config)

        if stage:
            return vsi and vxlan
        else:
            return [vsi, vxlan]

    def remove_vsi(self, stage=False, vsi=None):
        """Stage or execute a config for removing a VSI

        Args:
            stage (bool): whether to stage the command or execute immediately

        Returns:
            True if stage=True and successfully staged
            etree.Element XML response if immediately executed
        """
        vsi_config = self._build_vsi('delete', vsi)
        if stage:
            vsi = self.device.stage_config(vsi_config, 'edit_config')
        else:
            vsi = self.device.edit_config(vsi_config)

        return vsi

    def remove_vxlan(self, stage=False):
        """Stage or execute a config for removing a VXLAN

        Args:
            stage (bool): whether to stage the command or execute immediately

        Returns:
            True if stage=True and successfully staged
            etree.Element XML response if immediately executed
        """
        vlan_config = self._build_vxlan('delete')
        if stage:
            vlan = self.device.stage_config(vlan_config, 'edit_config')
        else:
            vlan = self.device.edit_config(vlan_config)

        return vlan

    def get_tunnels(self):
        """Get a list of tunnel interface that are mapped to a given VXLAN ID

        Returns:
            List of tunnel IDs

        """
        E = data_element_maker()
        top = E.top(
            E.VXLAN(
                E.Tunnels(
                    E.Tunnel(
                        E.VxlanID(self.vxlan)
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        tunnels_xml = findall_in_data('Tunnel', nc_get_reply.data_ele)

        tunnels = []
        for tunnel in tunnels_xml:
            tunnels.append(find_in_data('TunnelID', tunnel).text)

        return tunnels


class L2EthService(object):
    """This class is used to get data and configure Ethernet Service Instances
        on Layer 2 interfaces, map to VSI, and perform equiv of xconnect.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        interface (str): name of a Layer 2 interface
        instance (str): service instance ID to be configured on the
            interface
        vsi (str): name of the VSI being mapped to the instance

    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        interface (str): name of a Layer 2 interface
        instance (str): service instance ID to be configured on the
            interface
        vsi (str): name of the VSI being mapped to the instance

    """
    def __init__(self, device, interface, instance, vsi):

        self.device = device
        self.interface = interface
        self.vsi = vsi
        self.instance = instance

        # the next few attributes map dictionary keys to XML tags
        # and maps XML values to more user friendly values
        self.AC_KEY_MAP = {
            'vsi': 'VsiName',
            'index': 'IfIndex',
            'instance': 'SrvID',
            'access_mode': 'AccessMode',
        }
        self.AC_VALUE_MAP = {
            'AccessMode': {
                '1': 'vlan',
                '2': 'ethernet'
            }
        }
        self.RV_KEY_MAP = {
            'index': 'IfIndex',
            'instance': 'SrvID',
            'encap': 'Encap',
            'vlanid': 'SVlanRange',
            'cvid': 'CVlanRange'
        }
        self.RV_VALUE_MAP = {
            'Encap': {
                '1': 'default',
                '2': 'untagged',
                '3': 'tagged',
                '4': 's-vid',
                '5': 'only-tagged',
                '6': 'SvlanIdCvlanId',
                '7': 'CvlanId',
                '8': 'CvlanList',
                '9': 'SvlanIdCvlanList',
                '10': 'SvlanIdCvlanAll',
                '11': 'SvlanList',
                '12': 'SvlanListOnlyTagged',
            }
        }

        """
        supported value definition when using Ansible
        encapsulation default - 1
        encapsulation s-vid VLAN_ID - 4, but adds SVlanRange key as VLAN_ID
        encapsulation s-vid VLAN_ID only-tagged - 5,
                 but adds SVlanRange key as VLAN_ID
        encapsulation tagged - 3
        encapsulation untagged - 2
        """

    def get_config(self):
        """Get config of a service instance on a given interface for a
            given VSI

        Returns:
            If the mapping exists, it returns a dictionary with the
            following key/value pairs:
                :index (str): value of IfIndex
                :interface (str): name of interface
                :vsi (str): name of VSI
                :instance (str): instance ID
                :encap (str): ['default', 'untagged', 'tagged', 's-vid',
                    'only-tagged']
                :vlanid (str): vlanid PRESENT when ``encap`` set to
                    "only-tagged" or "s-vid"
                :access_mode (str): "vlan" or "ethernet"

        """
        existing = {}
        existing.update(self.get_vsi_encap())
        if existing:
            existing.update(self.get_vsi_map())
        return existing

    def vsi_exist(self):
        """Check to see if the VSI exists

        Returns:
            If returns True if the VSI exists, else false)

        """
        E = data_element_maker()
        top = E.top(
            E.L2VPN(
                E.VSIs(
                    E.VSI(
                        E.VsiName(self.vsi)
                    )
                )
            )
        )

        VSI = {
            'vsi': 'VsiName',
        }

        nc_get_reply = self.device.get(('subtree', top))
        return_map = data_elem_to_dict(nc_get_reply.data_ele, VSI)

        return return_map

    def get_vsi_map(self):
        """Get xconnect config for given interface and service instance

        Returns:
            If the mapping exists, it returns a dictionary with the
            following key/value pairs:
                :index (str): value of IfIndex
                :vsi (str): name of VSI
                :interface (str): name of interface
                :instance (str): instance ID
                :access_mode (str): "vlan" or "ethernet"

        """
        E = data_element_maker()
        top = E.top(
            E.L2VPN(
                E.ACs(
                    E.AC(
                        E.SrvID(self.instance),
                        E.IfIndex(self._index_from_interface(self.interface))
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        return_map = data_elem_to_dict(
            nc_get_reply.data_ele, self.AC_KEY_MAP, value_map=self.AC_VALUE_MAP)

        return return_map

    def get_vsi_encap(self):
        """Gets encap configuration for a given VXLAN ID

        Returns:
            If a config exists, it returns a dictionary with the
            following key/value pairs:
                :index (str): value of IfIndex
                :instance (str): instance ID
                :encap (str): ['default', 'untagged', 'tagged', 's-vid',
                    'only-tagged']
                :vlanid (str): vlanid PRESENT when ``encap`` set to
                    "only-tagged" or "s-vid"
        """
        E = data_element_maker()
        top = E.top(
            E.L2VPN(
                E.SRVs(
                    E.SRV(
                        E.SrvID(self.instance),
                        E.IfIndex(self._index_from_interface(self.interface))
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        return_vsi = data_elem_to_dict(
            nc_get_reply.data_ele, self.RV_KEY_MAP, value_map=self.RV_VALUE_MAP)

        if return_vsi:
            return_vsi['interface'] = self._get_interface_from_index(
                return_vsi.get('index'))

        return return_vsi

    def _build_encap(self, operation, stage=False, **kvargs):
        """Stage or execute encap (service instance) config object for an interface

        Args:
            state (str): "present" or "absent"
            kvargs: see Keyword Args
            stage (bool): whether to stage the command or execute immediately

        Keyword Args:
            encap (str): 'default', 'tagged', 'untagged', 'only-tagged',
                's-vid'
            vlanid (str): REQUIRED if encap is set to only-tagged or s-vid

        Note:
            when encap is set to only-tagged, it also ensures s-vid

        Returns:
            True if stage=True and successfully staged
            etree.Element XML responses if immediately executed
        """
        EN = nc_element_maker()
        EC = config_element_maker()

        encap_eles = []
        if operation == 'merge':

            REVERSE_RV_KEY_MAP = dict(reversed(
                item) for item in self.RV_KEY_MAP.items())

            REVERSE_RV_VALUE_MAP = reverse_value_map(
                REVERSE_RV_KEY_MAP, self.RV_VALUE_MAP)

            encap = kvargs.get('encap')
            vlanid = kvargs.get('vlanid')
            if encap:
                if encap in ['default', 'tagged', 'untagged']:
                    value = REVERSE_RV_VALUE_MAP.get('encap').get(
                        kvargs.get('encap'))
                    encap_eles.append(EC.Encap(value))
                elif encap in ['s-vid', 'only-tagged']:
                    encap_eles.append(EC.SVlanRange(vlanid))
                    if encap == 's-vid':
                        encap_eles.append(EC.Encap('4'))
                    elif encap == 'only-tagged':
                        encap_eles.append(EC.Encap('5'))

        self.jindex = self._index_from_interface(self.interface) # minor hack for now

        config = EN.config(
            EC.top(
                EC.L2VPN(
                    EC.SRVs(
                        EC.SRV(
                            EC.SrvID(self.instance),
                            EC.IfIndex(self.jindex),
                            *encap_eles
                        )
                    ),
                    **operation_kwarg(operation)
                )
            )
        )

        if stage:
            self.device.stage_config(config, 'edit_config')
        else:
            self.device.edit_config(config)

    def remove(self, stage=False):
        """Stage or execute object to remove service instance configuration

        Args:
            stage (bool): whether to stage the command or execute immediately
        """
        self._build_config(state='absent', stage=stage)

    def build(self, stage=False, **kvargs):
        """Builds service instance (and xconn) config object for an interface
        Args:
            state (str): "present" or "absent"
            kvargs: see Keyword Args

        Keyword Args:
            vsi (str): OPTIONAL - name of VSI
            instance (str): OPTIONAL - instance ID
            encap (str): REQUIRED - ['default', 'untagged', 'tagged',
                's-vid', 'only-tagged']
            vlanid (str): REQUIRED when ``encap`` set to
               "only-tagged" or "s-vid"
            access_mode (str): OPTIONAL - "vlan" or "ethernet"

        Note: when encap is set to only-tagged, it also ensures s-vid
        """
        return self._build_config(state='present', stage=stage, **kvargs)

    def _build_config(self, state, stage=False, **kvargs):
        """Stage or execute service instance (and xconn) config object for an interface

        Args:
            stage (bool): whether to stage the command or execute immediately
        """
        if state == 'present':
            operation = 'merge'
        elif state == 'absent':
            operation = 'delete'

        self._build_encap(operation, stage=stage, **kvargs)

        if operation == 'merge':
            # needs to happen AFTER
            self._build_xconnect(operation, self.jindex, stage=stage, **kvargs)

    def _build_xconnect(self, operation, index, stage=False, **kvargs):
        """Stage or execute config object to configure the equivalent of the xconnect
        command

        Args:
            operation (str): "merge" or "delete"
            index (str): IfIndex of the interface being configured
            stage (bool): whether to stage the command or execute immediately
            kvargs: see Keyword Args

        Keyword Args:
            access_mode (str): "vlan" or "ethernet"

        Returns:
            True if stage=True and successfully staged
            etree.Element XML responses if immediately executed
        """
        EN = nc_element_maker()
        EC = config_element_maker()

        MAP = {
            'vlan': '1',
            'ethernet': '2'
        }

        config = EN.config(
            EC.top(
                EC.L2VPN(
                    EC.ACs(
                        EC.AC(
                            EC.SrvID(self.instance),
                            EC.IfIndex(index),
                            EC.VsiName(self.vsi),
                            EC.AccessMode(MAP.get(kvargs.get('access_mode')))
                        )
                    )
                )
            )
        )

        if stage:
            self.device.stage_config(config, 'edit_config')
        else:
            self.device.edit_config(config)

    def _index_from_interface(self, interface):
        """Returns IfIndex from a given Interface name
        """
        intf = Interface(self.device, interface)
        return intf.iface_index

    def _get_interface_from_index(self, index):
        """ Returns interface name based on a given IfIndex
        """
        interface_name = None
        if index:
            E = data_element_maker()
            top = E.top(
                E.Ifmgr(
                    E.Interfaces(
                        E.Interface(
                            E.IfIndex(index),
                            E.Name()
                        )
                    )
                )
            )

            nc_get_reply = self.device.get(('subtree', top))
            interface_name = find_in_data('Name', nc_get_reply.data_ele).text

        return interface_name



"""Errors for pyhpecw7 library.

(c) Copyright 2016 Hewlett Packard Enterprise Development LP Licensed under the Apache License, Version 2.0
(the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License
at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
express or implied. See the License for the specific language governing permissions and limitations under the License.

"""
from pyhpecw7.utils.xml.lib import remove_namespaces, get_text


class PYHPError(Exception):
    """General base class for all errors in the pyhpecw7 library.
    """
    pass


class FeatureError(PYHPError):
    """Base class for all errors related to features.
    """
    pass

##################################
#       NETCONF RPC ERRORS       #
##################################


class NCError(PYHPError):
    """Wrapper for ncclient RPC errors.
    """
    def __init__(self, rpc_error=None):
        super(NCError, self).__init__()
        self.tag = ''
        if rpc_error:
            self.rpc_error = rpc_error
            self.xml = remove_namespaces(rpc_error.xml)
            self.msg = rpc_error.message
            self.tag = rpc_error.tag
            self.bad_element = get_text(self.xml, 'error-info/bad-element') or ''

    def __repr__(self):
        rep = '{0}: NETCONF Protocol error ({1}).'.format(self.__class__.__name__, self.msg)

        if hasattr(self, 'rpc_error'):
            rep += ' Tag: {0}'.format(self.tag)
            if self.bad_element:
                rep += ' Bad Element: {0}'.format(self.bad_element)

        if self.tag == 'access-denied':
            rep += ' It is possible the device is'\
                + ' locked by another NETCONF connection.'

        if self.tag == 'invalid-value':
            rep += ' Message: {0}'.format(self.msg)

        return rep

    __str__ = __repr__


class LockConflictError(PYHPError):
    """When there's an attempt to lock, and NETCONF lock is not available.
    """
    def __repr__(self):
        return '{0}: Couldn\'t obtain NETCONF lock.'.format(
            self.__class__.__name__) +\
            ' Another connection holds the lock.'

    __str__ = __repr__


class UnlockConflictError(PYHPError):
    """When there's an attempt to unlock, and NETCONF lock is not available.
    """
    def __repr__(self):
        return '{0}: Couldn\'t unlock NETCONF lock.'.format(
            self.__class__.__name__) +\
            ' Another connection holds the lock.'

    __str__ = __repr__

##################################
#       TIMEOUT ERRORS         #
##################################


class NCTimeoutError(PYHPError):
    """When there is no response from the device within the timeout range.
    """
    def __repr__(self):
        return '{0}: The NETCONF RPC timed out.'.format(
            self.__class__.__name__)

    __str__ = __repr__


##################################
#       CONNECTION ERRORS        #
##################################


class ConnectionError(PYHPError):
    """When there is an error in the SSH/NETCONF connection.
    """
    def __init__(self, dev, msg=None):
        super(ConnectionError, self).__init__()
        self.dev = dev
        self.msg = msg
        self.host = dev.host
        self.port = dev.port

    def __repr__(self):
        rep = '{0}: host: {1}, port: {2}'.format(
            self.__class__.__name__, self.host, self.port)
        if self.msg:
            rep += ' msg: {0}'.format(self.msg)

        return rep

    __str__ = __repr__


class ConnectionAuthenticationError(ConnectionError):
    """When there's an authentication error.
    """
    pass


class ConnectionSSHError(ConnectionError):
    """When there's an SSH error.
    """
    pass


class ConnectionUnkownHostError(ConnectionError):
    """When there's an uknown host error.
    """
    pass


class ConnectionClosedError(ConnectionError):
    """When there's a connection closed error.
    """
    pass



from .network import ipaddr


def valid_ip_network(network):
    """Take a v4 or v6 network, e.g. '192.168.3.5/24' or
    '192.168.3.5/255.255.255.0' and return whether it is valid.

    Args:
        network (str): IP address and mask, e.g. '192.168.3.5/24'.

    Returns:
        True if valid, False otherwise.
    """
    try:
        ipaddr.IPNetwork(network)
    except ValueError:
        return False

    return True
